import tempfile
import os.path
import re
import yaml
from dagshub.logger import dagshub_logger, DAGsHubLogger


def in_tmp_dir(f):
    with tempfile.TemporaryDirectory() as d:
        metrics_path = os.path.join(d, "metrics.csv")
        hparams_path = os.path.join(d, "hparams.yml")
        f(metrics_path, hparams_path)


def test_context_manager_no_eager_logging():
    def f(metrics_path, hparams_path):
        with dagshub_logger(
            metrics_path=metrics_path, hparams_path=hparams_path, eager_logging=False
        ) as logger:
            logger.log_metrics({"a": 1, "b": 2})
            logger.log_hyperparams({"R": 2, "D": 2})
            logger.log_metrics(a=3, c=42, step_num=2)
            logger.log_hyperparams(R=5, lr=1e-4)
            assert not os.path.exists(metrics_path)
            assert not os.path.exists(hparams_path)

        assert os.path.exists(metrics_path)
        assert os.path.exists(hparams_path)

        with open(metrics_path) as metrics_file:
            assert metrics_file.readline() == "Name,Value,Timestamp,Step\n"
            assert re.compile('^"a",1,\\d+,1\n$').match(metrics_file.readline())
            assert re.compile('^"b",2,\\d+,1\n$').match(metrics_file.readline())
            assert re.compile('^"a",3,\\d+,2\n$').match(metrics_file.readline())
            assert re.compile('^"c",42,\\d+,2\n$').match(metrics_file.readline())
            assert not metrics_file.readline()

        with open(hparams_path) as hparams_file:
            assert yaml.safe_load(hparams_file) == {"R": 5, "D": 2, "lr": 1e-4}

    in_tmp_dir(f)


def test_eager_logging():
    def f(metrics_path, hparams_path):
        logger = DAGsHubLogger(
            metrics_path=metrics_path, hparams_path=hparams_path, eager_logging=True
        )
        assert os.path.exists(metrics_path)
        assert os.path.exists(hparams_path)
        logger.close()

    in_tmp_dir(f)


def test_forbidden_csv_chars_in_metric_names():
    def f(metrics_path, hparams_path):
        with dagshub_logger(
            metrics_path=metrics_path, hparams_path=hparams_path
        ) as logger:
            logger.log_metrics(
                {
                    "this/is/forbidden": 1,
                    "so,is,this": 2,
                    'and "this"': 3,
                    "also \n this": 4,
                    "normal": 5,
                }
            )

        lines = list(open(metrics_path))
        assert lines[0] == "Name,Value,Timestamp,Step\n"
        assert lines[1].startswith('"this/is/forbidden",1'), lines[1]
        assert lines[2].startswith('"so,is,this",2'), lines[2]
        assert lines[3].startswith('"and ""this""",3'), lines[3]
        assert lines[4] == '"also \n'
        assert lines[5].startswith(' this",4'), lines[5]
        assert lines[6].startswith('"normal",5'), lines[6]

    in_tmp_dir(f)


if __name__ == "__main__":
    test_forbidden_csv_chars_in_metric_names()
    test_context_manager_no_eager_logging()
    test_eager_logging()



import inspect
import logging
import multiprocessing.pool
from dataclasses import field, dataclass
from itertools import repeat
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING, List, Dict, Any, Optional, Union

import rich.progress

from dagshub.common.analytics import send_analytics_event
from dagshub.common.download import download_files
from dagshub.common.helpers import sizeof_fmt, prompt_user
from dagshub.common.rich_util import get_rich_progress
from dagshub.common.util import lazy_load
from dagshub.data_engine.annotation.voxel_conversion import add_voxel_annotations, add_ls_annotations
from dagshub.data_engine.client.models import DatasourceType
from dagshub.data_engine.model.datapoint import Datapoint, _get_blob
from dagshub.data_engine.client.loaders.base import DagsHubDataset
from dagshub.data_engine.voxel_plugin_server.utils import set_voxel_envvars

if TYPE_CHECKING:
    from dagshub.data_engine.model.datasource import Datasource
    import fiftyone as fo
    import dagshub.data_engine.voxel_plugin_server.server as plugin_server_module
else:
    plugin_server_module = lazy_load("dagshub.data_engine.voxel_plugin_server.server")
    fo = lazy_load("fiftyone")
    tf = lazy_load("tensorflow")

logger = logging.getLogger(__name__)


class VisualizeError(Exception):
    pass


@dataclass
class QueryResult:
    _entries: List[Datapoint]
    """ List of downloaded entries."""
    datasource: "Datasource"
    _datapoint_path_lookup: Dict[str, Datapoint] = field(init=False)

    def __post_init__(self):
        self._refresh_lookups()

    @property
    def entries(self):
        return self._entries

    @entries.setter
    def entries(self, value: List[Datapoint]):
        self._entries = value
        self._refresh_lookups()

    def _refresh_lookups(self):
        self._datapoint_path_lookup = {}
        for e in self.entries:
            self._datapoint_path_lookup[e.path] = e

    @property
    def dataframe(self):
        import pandas as pd

        metadata_keys = set()
        for e in self.entries:
            metadata_keys.update(e.metadata.keys())

        metadata_keys = list(sorted(metadata_keys))
        return pd.DataFrame.from_records(
            [dp.to_dict(metadata_keys) for dp in self.entries]
        )

    def __len__(self):
        return len(self.entries)

    def __iter__(self):
        return self.entries.__iter__()

    def __repr__(self):
        return f"QueryResult of datasource {self.datasource.source.name} with {len(self.entries)} datapoint(s)"

    @staticmethod
    def from_gql_query(
        query_resp: Dict[str, Any], datasource: "Datasource"
    ) -> "QueryResult":
        if "edges" not in query_resp:
            return QueryResult([], datasource)
        if query_resp["edges"] is None:
            return QueryResult([], datasource)
        return QueryResult(
            [Datapoint.from_gql_edge(edge, datasource) for edge in query_resp["edges"]],
            datasource,
        )

    def as_ml_dataset(self, flavor, **kwargs):
        """
        Convert the QueryResult into a dataset for a machine learning framework

        ARGS:
        flavor: torch|tensorflow

        KWARGS:
        metadata_columns: columns that are returned from the metadata as part of the dataset
        strategy: preload|background|lazy; default: lazy
        savedir: location at which the dataset is stored
        processes: number of parallel processes that download the dataset
        tensorizer: auto|image|<function>
        """
        flavor = flavor.lower()
        if flavor == "torch":
            from dagshub.data_engine.client.loaders.torch import PyTorchDataset

            return PyTorchDataset(self, **kwargs)
        elif flavor == "tensorflow":
            from dagshub.data_engine.client.loaders.tf import TensorFlowDataset

            ds_builder = TensorFlowDataset(self, **kwargs)
            ds = tf.data.Dataset.from_generator(
                ds_builder.generator, output_signature=ds_builder.signature
            )
            ds.__len__ = lambda: ds_builder.__len__()
            ds.__getitem__ = ds_builder.__getitem__
            ds.builder = ds_builder
            ds.type = "tensorflow"
            return ds
        else:
            raise ValueError("supported flavors are torch|tensorflow")

    def as_ml_dataloader(self, flavor, **kwargs):
        """
        Convert the QueryResult into a dataloader for a machine learning framework

        ARGS:
        flavor: torch|tensorflow

        KWARGS:
        metadata_columns: columns that are returned from the metadata as part of the dataloader
        strategy: preload|background|lazy; default: lazy
        savedir: location at which the dataset is stored
        processes: number of parallel processes that download the dataset
        tensorizer: auto|image|<function>
        for_dataloader: bool; internal argument, that begins background dataset download after
                        the shuffle order is determined for the first epoch; default: False
        """

        send_analytics_event("Client_DataEngine_DataLoaderInitialized", repo=self.datasource.source.repoApi)

        def keypairs(keys):
            return {key: kwargs[key] for key in keys}

        if type(flavor) != str:
            if flavor.type == "torch":
                from dagshub.data_engine.client.loaders.torch import PyTorchDataLoader

                return PyTorchDataLoader(flavor, **kwargs)
            elif flavor.type == "tensorflow":
                from dagshub.data_engine.client.loaders.tf import TensorFlowDataLoader

                return TensorFlowDataLoader(flavor, **kwargs)

        kwargs["for_dataloader"] = True
        dataset_kwargs = set(
            list(inspect.signature(DagsHubDataset).parameters.keys())[1:]
        )
        global_kwargs = set(kwargs.keys())
        flavor = flavor.lower() if type(flavor) == str else flavor
        if flavor == "torch":
            from dagshub.data_engine.client.loaders.torch import PyTorchDataLoader

            return PyTorchDataLoader(
                self.as_ml_dataset(
                    flavor, **keypairs(global_kwargs.intersection(dataset_kwargs))
                ),
                **keypairs(global_kwargs - dataset_kwargs),
            )
        elif flavor == "tensorflow":
            from dagshub.data_engine.client.loaders.tf import TensorFlowDataLoader

            return TensorFlowDataLoader(
                self.as_ml_dataset(
                    flavor,
                    **keypairs(global_kwargs.intersection(dataset_kwargs)),
                ),
                **keypairs(global_kwargs - dataset_kwargs),
            )
        else:
            raise ValueError(
                "supported flavors are torch|tensorflow|<torch.utils.data.Dataset>|<tf.data.Dataset>"
            )

    def __getitem__(self, item: Union[str, int, slice]):
        """
        Gets datapoint by its path (string) or by its index in the result (or slice)
        """
        if type(item) is str:
            return self._datapoint_path_lookup[item]
        elif type(item) is int:
            return self.entries[item]
        elif type(item) is slice:
            return QueryResult(_entries=self.entries[item], datasource=self.datasource)
        else:
            raise ValueError(
                f"Can't lookup datapoint using value {item} of type {type(item)}, "
                f"needs to be either int or str or slice")

    def get_blob_fields(self, *fields: str, load_into_memory=False,
                        cache_on_disk=True, num_proc: int = 32) -> "QueryResult":
        """
        Downloads data from blob fields

        Args:
            fields: list of binary fields to download blobs for
            load_into_memory: Whether to load the blobs into the datapoints, or just store them on disk
                If True : the datapoints' specified fields will contain the blob data
                If False: the datapoints' specified fields will contain Path objects to the file of the downloaded blob
            cache_on_disk: Whether to cache the blobs on disk or not (valid only if load_into_memory is set to True)
                Cache location is `~/dagshub/datasets/<user>/<repo>/<datasource_id>/.metadata_blobs/`
        """
        if not load_into_memory:
            assert cache_on_disk
        for fld in fields:
            logger.info(f"Downloading metadata for field {fld} with {num_proc} processes")
            cache_location = self.datasource.default_dataset_location / ".metadata_blobs"

            if cache_on_disk:
                cache_location.mkdir(parents=True, exist_ok=True)

            blob_urls = list(map(lambda dp: dp._extract_blob_url_and_path(fld), self.entries))
            auth = self.datasource.source.repoApi.auth
            func_args = zip(map(lambda bu: bu[0], blob_urls), map(lambda bu: bu[1], blob_urls), repeat(auth),
                            repeat(cache_on_disk), repeat(load_into_memory))
            with multiprocessing.pool.ThreadPool(num_proc) as pool:
                res = pool.starmap(_get_blob, func_args)

            for dp, binary_val_or_path in zip(self.entries, res):
                if binary_val_or_path is None:
                    continue
                dp.metadata[fld] = binary_val_or_path

        return self

    def download_binary_columns(self, *columns: str, load_into_memory=True,
                                cache_on_disk=True, num_proc: int = 32) -> "QueryResult":
        """
        deprecated: Use get_blob_columns instead.
        """
        send_analytics_event("Client_DataEngine_downloadBlobs", repo=self.datasource.source.repoApi)
        return self.get_blob_fields(*columns, load_into_memory=load_into_memory, cache_on_disk=cache_on_disk,
                                    num_proc=num_proc)

    def download_files(self, target_dir: Optional[Union[str, PathLike]] = None, keep_source_prefix=True,
                       redownload=False, path_field: Optional[str] = None) -> PathLike:
        """
        Downloads the query result to the target_dir directory
        Args:
            target_dir: Where to download the files.
                If not specified, then downloads to ~/dagshub/datasets/<user>/<repo>/<ds_id>
            keep_source_prefix: If True, includes the prefix of the datasource in the download path
                Useful for cases where the download path is the root of the repository
            redownload: Whether to redownload a file if it exists on the filesystem already
                NOTE: We don't do any hashsum checks, so if it's possible that the file has been updated, turn it on
            path_field: If you want to download files from a field other than the datapoint's path.
                NOTE: the path still needs to be in the same datasource
                and have the same format as the path of the datapoint,
                for now you can't download arbitrary paths/urls
        Returns:
            Path to the directory with the downloaded files
        """

        target_path = self.datasource.default_dataset_location if target_dir is None else Path(target_dir).expanduser()
        logger.warning(f"Downloading {len(self.entries)} files to {str(target_path)}")

        if self.datasource.source.source_type == DatasourceType.BUCKET:
            send_analytics_event("Client_DataEngine_downloadDatapointsFromBucket", repo=self.datasource.source.repoApi)
        else:
            send_analytics_event("Client_DataEngine_downloadDatapoints", repo=self.datasource.source.repoApi)

        def dp_path(dp: Datapoint):
            if path_field is not None:
                path_val = dp.metadata.get(path_field)
                if path_val is None:
                    return None
            else:
                path_val = dp.path

            if keep_source_prefix:
                return target_path / self.datasource.source.source_prefix / path_val
            else:
                return target_path / path_val

        def dp_url(dp: Datapoint):
            if path_field is not None:
                path_val = dp.metadata.get(path_field)
                if path_val is None:
                    return None
                return self.datasource.source.raw_path(path_val)
            else:
                return dp.download_url

        download_args = [(dp_url(dp), dp_path(dp)) for dp in self.entries if dp_path(dp) is not None]

        download_files(download_args, skip_if_exists=not redownload)
        return target_path

    def to_voxel51_dataset(self, **kwargs) -> "fo.Dataset":
        """
        Creates a voxel51 dataset that can be used with `fo.launch_app()` to run it

        Kwargs:
            name (str): name of the dataset (by default uses the same name as the datasource)
            force_download (bool): download the dataset even if the size of the files is bigger than 100MB
            files_location (str|PathLike): path to the location where to download the local files
                default: ~/dagshub_datasets/user/repo/ds_name/
            redownload (bool): Redownload files, replacing the ones that might exist on the filesystem
            voxel_annotations (List[str]) : List of fields from which to load voxel annotations serialized with
                                        `to_json()`. This will override the labelstudio annotations
        """
        if len(self.entries) == 0:
            raise VisualizeError("No datapoints to visualize")
        logger.info("Migrating dataset to voxel51")
        name = kwargs.get("name", self.datasource.source.name)
        force_download = kwargs.get("force_download", False)
        # TODO: don't override samples in existing one
        if fo.dataset_exists(name):
            ds: fo.Dataset = fo.load_dataset(name)
        else:
            ds: fo.Dataset = fo.Dataset(name)
        # ds.persistent = True

        dataset_location = Path(kwargs.get("files_location", self.datasource.default_dataset_location))
        dataset_location.mkdir(parents=True, exist_ok=True)
        logger.info("Downloading files...")

        if "voxel_annotations" in kwargs:
            annotation_fields = kwargs["voxel_annotations"]
            label_func = add_voxel_annotations
        else:
            annotation_fields = self.datasource.annotation_fields
            label_func = add_ls_annotations

        # Load the annotation fields
        self.download_binary_columns(*annotation_fields)

        if not force_download:
            self._check_downloaded_dataset_size()

        redownload = kwargs.get("redownload", False)
        self.download_files(self.datasource.default_dataset_location, redownload=redownload)

        progress = get_rich_progress(rich.progress.MofNCompleteColumn())
        task = progress.add_task("Generating voxel samples...", total=len(self.entries))

        samples: List["fo.Sample"] = []

        with progress:
            for datapoint in self.entries:
                filepath = self.datasource.default_dataset_location / datapoint.path_in_repo
                sample = fo.Sample(filepath=filepath)
                sample["dagshub_download_url"] = datapoint.download_url
                sample["datapoint_id"] = datapoint.datapoint_id
                sample["datapoint_path"] = datapoint.path
                label_func(sample, datapoint, *annotation_fields)
                for k, v in datapoint.metadata.items():
                    # TODO: more filtering here, not all fields should be showing up in voxel
                    if k in annotation_fields:
                        continue
                    if type(v) is not bytes:
                        sample[k] = v
                samples.append(sample)
                progress.update(task, advance=1, refresh=True)

        ds.merge_samples(samples)
        return ds

    def _check_downloaded_dataset_size(self):
        download_size_prompt_threshold = 100 * (2 ** 20)  # 100 Megabytes
        dp_size = self._calculate_datapoint_size()
        if dp_size is not None and dp_size > download_size_prompt_threshold:
            prompt = f"You're about to download {sizeof_fmt(dp_size)} of images locally."
            should_download = prompt_user(prompt)
            if not should_download:
                msg = "Downloading voxel dataset cancelled"
                logger.warning(msg)
                raise RuntimeError(msg)

    def _calculate_datapoint_size(self) -> Optional[int]:
        sum_size = 0
        has_sum_field = False
        all_have_sum_field = True
        size_field = "size"
        for dp in self.entries:
            if size_field in dp.metadata:
                has_sum_field = True
                sum_size += dp.metadata[size_field]
            else:
                all_have_sum_field = False
        if not has_sum_field:
            logger.warning("None of the datapoints had a size field, can't calculate size of the downloading dataset")
            return None
        if not all_have_sum_field:
            logger.warning("Not every datapoint has a size field, size calculations might be wrong")
        return sum_size

    def visualize(self, **kwargs):
        set_voxel_envvars()

        send_analytics_event("Client_DataEngine_VizualizeResults", repo=self.datasource.source.repoApi)

        ds = self.to_voxel51_dataset(**kwargs)

        sess = fo.launch_app(ds)
        # Launch the server for plugin interaction
        plugin_server_module.run_plugin_server(sess, self.datasource, self.datasource.source.revision)

        return sess

    def annotate(self, open_project=True, ignore_warning=True) -> Optional[str]:
        """
        Sends all the datapoints returned in this QueryResult to be annotated in Label Studio on DagsHub.

        :param open_project: Whether to automatically open the returned URL in your browser
        :param ignore_warning: Suppress any non-lethal warnings that require user input
        :return The URL of the created Label Studio workspace
        """
        send_analytics_event("Client_DataEngine_SentToAnnotation", repo=self.datasource.source.repoApi)

        return self.datasource.send_datapoints_to_annotation(self.entries,
                                                             open_project=open_project, ignore_warning=ignore_warning)



import logging
from typing import TYPE_CHECKING, List

import dacite
from starlette.requests import Request
from starlette.responses import JSONResponse

from dagshub.data_engine.model.query import DatasourceQuery
from dagshub.data_engine.voxel_plugin_server.models import VoxelFilterState
from dagshub.data_engine.voxel_plugin_server.routes.util import error_handler
from dagshub.data_engine.voxel_plugin_server.utils import get_plugin_state

if TYPE_CHECKING:
    from dagshub.data_engine.model.datasource import Datasource
    import fiftyone as fo

logger = logging.getLogger(__name__)


@error_handler
async def save_dataset(request: Request):
    plugin_state = get_plugin_state(request)
    data = await request.json()
    name = data["name"]
    if data["saveVoxelFilters"]:
        filters = await get_voxel_filters(request)
        ds_with_filters = apply_filters_to_datasource(plugin_state.datasource, filters)
    else:
        ds_with_filters = plugin_state.datasource

    try:
        ds_with_filters.save_dataset(name)
        return JSONResponse(f"Dataset {name} saved successfully")
    except Exception as e:
        return JSONResponse(f"Error while saving dataset: {e}", status_code=400)


async def get_voxel_filters(request: Request) -> List[VoxelFilterState]:
    plugin_state = get_plugin_state(request)
    filter_dict = (await request.json())["filters"]

    def transform(field, filter_dict):
        filter = dacite.from_dict(VoxelFilterState, filter_dict)
        filter.filter_field = field
        return filter

    filters = [transform(k, v) for k, v in filter_dict.items()]
    filters = sanitize_voxel_filter(plugin_state.voxel_session, filters)
    return filters


_generated_fields = ["dagshub_download_url", "filepath", "id", "datapoint_id", "datapoint_path"]


def sanitize_voxel_filter(sess: "fo.Session", filters: List[VoxelFilterState]):
    """
    This command sanitizes the voxel filter dictionary.
    Included sanitization:
        - Range queries: check for min/max value in the range and remove the bounds
            if e.g. the min value is the same as the dataset's min value
        - Removes reserved/generated fields (like dagshub_download_url)
    """
    dataset = sess.dataset
    res_filters = []
    for f in filters:
        if f.filter_field in _generated_fields:
            logger.warning(f"Not applying filter to field {f.filter_field} because it's a generated field")
        else:
            # TODO: maybe get the metadata from the datasource and only filter for existing fields
            # Remove bounds for range queries
            if f.range is not None:
                min_val, max_val = dataset.bounds(f.filter_field)
                if f.range[0] == min_val:
                    f.range[0] = None
                if f.range[1] == max_val:
                    f.range[1] = None
                if f.range == [None, None]:
                    # Do not add if there are no bounds
                    continue
            res_filters.append(f)
    return res_filters


def apply_filters_to_datasource(ds: "Datasource", filters: List[VoxelFilterState]) -> "Datasource":
    voxel_query = DatasourceQuery()
    for f in filters:
        filter_query = f.to_datasource_query()
        voxel_query.compose("and", filter_query)
    return ds.add_query_op("and", voxel_query)



import functools
from typing import Optional, Any, Dict, Union

from dagshub.data_engine.client.query_builder import GqlQuery


class GqlQueries:

    @staticmethod
    @functools.lru_cache()
    def datasource() -> str:
        q = GqlQuery().operation(
            "query",
            name="datasource",
            input={
                "$id": "ID",
                "$name": "String"
            }
        ).query(
            "datasource",
            input={
                "id": "$id",
                "name": "$name",
            }
        ).fields([
            "id",
            "name",
            "rootUrl",
            "integrationStatus",
            "preprocessingStatus",
            "metadataFields {name valueType multiple}"
            "type",
        ]).generate()
        return q

    @staticmethod
    def datasource_params(id: Optional[Union[int, str]], name: Optional[str]) -> Dict[str, Any]:
        return {
            "id": id,
            "name": name,
        }

    @staticmethod
    @functools.lru_cache()
    def datasource_query(include_metadata: bool) -> str:
        metadata_fields = "metadata { key value }" if include_metadata else ""
        q = GqlQuery().operation(
            "query",
            name="datasourceQuery",
            input={
                "$datasource": "ID!",
                "$queryInput": "QueryInput",
                "$first": "Int",
                "$after": "String",
            }
        ).query(
            "datasourceQuery",
            input={
                "datasource": "$datasource",
                "filter": "$queryInput",
                "first": "$first",
                "after": "$after",
            }
        ).fields([
            f"edges {{ node {{ id path {metadata_fields} }} }}",
            "pageInfo { hasNextPage endCursor }",
        ]).generate()
        return q

    @staticmethod
    def datasource_query_params(datasource_id: Union[int, str], query_input: Dict[str, Any], first: Optional[int],
                                after: Optional[str]) -> Dict[str, Any]:
        return {
            "datasource": datasource_id,
            "queryInput": query_input,
            "first": first,
            "after": after,
        }

    @staticmethod
    @functools.lru_cache()
    def dataset() -> str:
        q = GqlQuery().operation(
            "query",
            name="dataset",
            input={
                "$id": "ID",
                "$name": "String"
            }
        ).query(
            "dataset",
            input={
                "dataset": "$id",
                "name": "$name",
            }
        ).fields([
            "id",
            "name",
            "datasource {id name rootUrl integrationStatus preprocessingStatus "
            "metadataFields {name valueType multiple} type}",
            "datasetQuery",
        ]).generate()
        return q

    @staticmethod
    def dataset_params(id: Optional[Union[int, str]], name: Optional[str]) -> Dict[str, Any]:
        return {
            "id": id,
            "name": name,
        }



import logging
import os.path
import signal
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import partial
from pathlib import Path
from typing import Tuple, Callable, Optional, List, Union, Dict

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import re
import rich.progress

from dagshub.auth import get_token
from dagshub.auth.token_auth import HTTPBearerAuth
from dagshub.common import config
from dagshub.common.helpers import http_request
from dagshub.common.rich_util import get_rich_progress

logger = logging.getLogger(__name__)

DownloadFunctionType = Callable[[str, Path], None]

storage_download_url_regex = re.compile(
    r".*/api/v1/repos/(?P<user>[\w\-_.]+)/(?P<repo>[\w\-_.]+)/storage/raw/(?P<proto>s3|gs)/"
    r"(?P<bucket>[a-z0-9.-]+)/(?P<path>.*)")


def enable_gcs_bucket_downloader(client=None):
    """
    Enables downloading storage items using the client, instead of going through DagsHub's server.
    For custom clients use `enable_custom_bucket_downloader` function

    Args:
        client: a google.cloud.storage.Client from the `google-cloud-storage` package.
            If client isn't specified, the default parameterless constructor is used
    """
    if client is None:
        from google.cloud import storage
        client = storage.Client()

    def get_fn(bucket_name, bucket_path) -> bytes:
        bucket = client.bucket(bucket_name)
        blob = bucket.blob(bucket_path).download_as_bytes()
        return blob

    add_bucket_downloader("gs", get_fn)


def enable_s3_bucket_downloader(client=None):
    """
    Enables downloading storage items using the client, instead of going through DagsHub's server.
    For custom clients use `enable_custom_bucket_downloader` function

    Args:
        client: a boto3 S3 client.
            If client isn't specified, the default parameterless constructor is used
    """

    if client is None:
        import boto3
        client = boto3.client("s3")

    def get_fn(bucket, path) -> bytes:
        resp = client.get_object(Bucket=bucket, Key=path)
        return resp["Body"].read()

    add_bucket_downloader("s3", get_fn)


def download_url_to_bucket_path(url: str) -> Optional[Tuple[str, str, str]]:
    """
    Gets a storage download URL of a dagshub file, returns a tuple of (protocol, bucket, path_in_bucket)
    """
    matches = storage_download_url_regex.match(url)
    if matches is None:
        return None
    groups = matches.groupdict()
    return groups["proto"], groups["bucket"], groups["path"]


def _dagshub_download(url: str, auth: HTTPBearerAuth) -> bytes:
    resp = http_request("GET", url, auth=auth, timeout=600)
    try:
        assert resp.status_code == 200
    # TODO: retry
    except AssertionError:
        raise RuntimeError(
            f"Couldn't download file at URL {url}. Response code {resp.status_code} (Body: {resp.content})")
    return resp.content


BucketDownloaderFuncType = Callable[[str, str], bytes]

_bucket_downloader_map: Dict[str, BucketDownloaderFuncType] = {}
_default_downloader: Optional[Callable[[str], bytes]] = None


def add_bucket_downloader(proto: Literal["gs", "s3"], func: BucketDownloaderFuncType):
    if proto in _bucket_downloader_map:
        logger.warning(f"Protocol {proto} already has a custom downloader function specified, overwriting it")
    _bucket_downloader_map[proto] = func


def _download_wrapper(url: str, location: Path, skip_if_exists: bool):
    if skip_if_exists and os.path.exists(location):
        return

    # Download the file
    # Check if it's a bucket
    bucket_tuple = download_url_to_bucket_path(url)
    assert _default_downloader is not None
    if bucket_tuple is None:
        # Not a bucket path - download regularly
        content = _default_downloader(url)
    else:
        # Bucket path - try to look if there's a custom downloader
        proto, bucket_name, bucket_path = bucket_tuple
        bucket_downloader = _bucket_downloader_map.get(proto)
        if bucket_downloader is None:
            content = _default_downloader(url)
        else:
            content = bucket_downloader(bucket_name, bucket_path)

    location.parent.mkdir(parents=True, exist_ok=True)
    with open(location, "wb") as f:
        f.write(content)


def _ensure_default_downloader_exists():
    """
    Checks that the default dagshub download function exists and prepares it otherwise
    """
    global _default_downloader
    if _default_downloader is None:
        token = config.token or get_token(host=config.host)
        _default_downloader = partial(_dagshub_download, auth=HTTPBearerAuth(token=token))


def download_files(files: List[Tuple[str, Union[str, Path]]],
                   download_fn: Optional[DownloadFunctionType] = None,
                   threads=32, skip_if_exists=True):
    """
    Download files using multithreading

    Parameters:
        files: list of (download_url: str, file_location: str or Path)
        download_fn: Optional function that will download the file. Needs to receive the two arguments:
            download url and Path where to save the file
            If function is not specified, then a default function that downloads a file with DagsHub credentials is used
            CAUTION: function needs to be pickleable since we're using ThreadPool to execute
        threads: number of threads to run this function on, default 32
        skip_if_exists: skip the download if the file exists (only for the default downloader)
    """
    _ensure_default_downloader_exists()

    # Convert string paths to Path objects
    for i, file_tuple in enumerate(files):
        if type(file_tuple[1]) is str:
            files[i] = (file_tuple[0], Path(file_tuple[1]))

    if download_fn is None:
        download_fn = partial(_download_wrapper, skip_if_exists=skip_if_exists)

    if len(files) > 1:
        # Multiple files - multithreaded download
        progress = get_rich_progress(rich.progress.MofNCompleteColumn(), transient=False)
        task = progress.add_task("Downloading files...", total=len(files))

        with progress:
            with ThreadPoolExecutor(max_workers=threads) as tp:

                def cancel_download(*args):
                    logger.warning("Interrupt received - shutting down downloader")
                    tp.shutdown(wait=False, cancel_futures=True)

                orig_interrupt = None
                try:
                    orig_interrupt = signal.signal(signal.SIGINT, cancel_download)
                # ValueError means the function is not running from the main thread.
                # TODO: figure out a workaround
                except ValueError:
                    pass

                futures = [tp.submit(download_fn, url, location) for (url, location) in files]
                for f in as_completed(futures):
                    exc = f.exception()
                    if exc is not None:
                        logger.warning(f"Got exception {type(exc)} while downloading file: {exc}")
                    progress.update(task, advance=1)

                if orig_interrupt is not None:
                    signal.signal(signal.SIGINT, orig_interrupt)

    elif len(files) == 1:
        # Single file - don't bother with the multithreading, just download the file
        url, location = files[0]
        try:
            download_fn(url, location)
        except Exception as exc:
            logger.warning(f"Got exception {type(exc)} while downloading file: {exc}")



import fnmatch
import json
import logging
import os
import posixpath
import time
import urllib
from http import HTTPStatus
from io import IOBase
from pathlib import Path
from typing import Union, Tuple, BinaryIO, Dict, Optional, Any, List

import httpx
import rich.progress
import rich.status

import dagshub.auth
from dagshub.auth.token_auth import HTTPBearerAuth
from dagshub.common import config, rich_console
from dagshub.common.api.repo import RepoAPI, BranchNotFoundError
from dagshub.common.helpers import log_message
from dagshub.upload.errors import determine_upload_api_error

# todo: handle api urls in common package
CONTENT_UPLOAD_URL = "api/v1/repos/{owner}/{reponame}/content/{branch}/{path}"
FILES_UI_URL = "{owner}/{reponame}/src/{branch}/{path}"
DEFAULT_COMMIT_MESSAGE = "Upload files using DagsHub client"
DEFAULT_DATASET_COMMIT_MESSAGE = "Initial dataset commit"
REPO_CREATE_URL = "api/v1/user/repos"
ORG_REPO_CREATE_URL = "api/v1/org/{orgname}/repos"
USER_INFO_URL = "api/v1/user"
DEFAULT_DATA_DIR_NAME = 'data'
logger = logging.getLogger(__name__)

s = httpx.Client()
s.timeout = config.http_timeout
s.follow_redirects = True
s.headers.update(config.requests_headers)

FileUploadStruct = Tuple[os.PathLike, BinaryIO]


def create_dataset(repo_name, local_path, glob_exclude="", org_name="", private=False):
    """
    Create a new repository on DagsHub and upload an entire dataset to it
    :param repo_name (str): Name of the repository to be created
    :param local_path (str): local path where the dataset to upload is located
    :param glob_exclude (str): regex to exclude certain files from the upload process
    :param org_name (str): Organization name to be the repository owner
    :param private (bool): Flag to indicate the repository is going to be private
    :return : Repo object of the repository created
    """
    repo = create_repo(repo_name, org_name=org_name, private=private)
    dir = repo.directory(repo_name)
    dir.add_dir(local_path, glob_exclude, commit_message=DEFAULT_DATASET_COMMIT_MESSAGE)
    return repo


def add_dataset_to_repo(repo,
                        local_path,
                        data_dir=DEFAULT_DATA_DIR_NAME):
    """
    Given a repository created on dagshub - upload an entire dataset to it
    :param reo (Reop): repository created beforehand
    :param local_path (str): local path where the dataset to upload is located
    :param data_dir (str): name of data directory that will be created inside repo
    """
    dir = repo.directory(data_dir)
    dir.add_dir(local_path, commit_message=DEFAULT_DATASET_COMMIT_MESSAGE)


def create_repo(
    repo_name,
    org_name="",
    description="",
    private=False,
    auto_init=False,
    gitignores="Python",
    license="",
    readme="",
    template="custom",
    host=""
):
    """
    Creates a repository on DagsHub for the current user (default) or an organization passed as an argument

    :param repo_name (str): Name of the repository to be created
    :param org_name (str): Organization name to be the repository owner
    :param description (str): Description for the repository
    :param private (bool): Flag to indicate the repository is going to be private
    :param auto_init (bool): Pass true to create an initial commit with README, .gitignore and LICENSE.
    :param gitignores (str): Which gitignore template(s) to use (comma separated string)
    :param license (str): Which license file to use
    :param readme (str): Readme file path to upload
    :param template (str): Which project template to use, options are: none, custom, notebook-template,
    cookiecutter-dagshub-dvc. To learn more, check out https://dagshub.com/docs/feature_guide/project_templates/
    :return: Repo object of the repository created
    """
    if template == "":
        template = "none"

    import dagshub.auth
    from dagshub.auth.token_auth import HTTPBearerAuth

    host = host or config.host

    username = config.username
    password = config.password
    if username is not None and password is not None:
        auth = username, password
    else:
        token = config.token or dagshub.auth.get_token()
        if token is not None:
            auth = HTTPBearerAuth(token)

    if auth is None:
        raise RuntimeError("You can't create a repository without being authenticated.")

    if (license != "" or readme != "" or gitignores != "") and template == "none":
        template = "custom"

    data = {
        "name": repo_name,
        "description": description,
        "private": private,
        "auto_init": auto_init,
        "gitignores": gitignores,
        "license": license,
        "readme": readme,
        "project_template": template,
    }

    url = REPO_CREATE_URL
    if org_name and not org_name.isspace():
        url = ORG_REPO_CREATE_URL.format(
            orgname=org_name,
        )

    res = s.post(urllib.parse.urljoin(host, url), data=data, auth=auth)

    if res.status_code != HTTPStatus.CREATED:
        logger.error(f"Response ({res.status_code}):\n" f"{res.content}")
        if res.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
            raise RuntimeError("Repository name is invalid or it already exists.")
        else:
            raise RuntimeError("Failed to create the desired repository.")

    repo = res.json()
    return Repo(
        owner=repo["owner"]["login"], name=repo["name"], token=token, branch="main"
    )


def validate_owner_repo(owner_repo: str) -> Tuple[str, str]:
    parts = owner_repo.split("/")
    if len(parts) != 2:
        raise ValueError("repo needs to be in the format <repo-owner>/<repo-name>")
    return parts[0], parts[1]


def upload_files(
    repo: str,
    local_path: Union[str, IOBase],
    commit_message=DEFAULT_COMMIT_MESSAGE,
    remote_path: str = None,
    **kwargs,
):
    """
    Convenience wrapper around Repo.upload
    :param repo: Repo identifier in the form <username>/<reponame>
    :param local_path: Specify the file or directory to be uploaded
    :param commit_message: Specify an optional commit message
    :param remote_path: Specify the path to upload the file to. Defaults to the relative path to the local_path.
    :param kwargs: Pass in any additional parameters that are required for the upload function
    """
    owner, repo = validate_owner_repo(repo)
    repo = Repo(owner, repo)
    repo.upload(local_path, commit_message=commit_message, remote_path=remote_path, **kwargs)


class Repo:
    def __init__(
        self, owner, name, username=None, password=None, token=None, branch=None
    ):

        """
        Repo class constructor. If branch is not provided, then default branch is used.

        WARNING: this class is not thread safe.
        Uploading files in parallel can lead to unexpected outcomes

        :param owner (str): Store the username of the user who owns this repository
        :param name (str): Identify the repository
        :param username (str): Set the username to none if it is not provided
        :param password (str): Set the password to none if it is not provided
        :param token (str): Set the token
        :param branch (str): Set the branch to the default branch
        :return: The object of the class

        """
        self.owner = owner
        self.name = name
        self.host = config.host

        self.username = username or config.username
        self.password = password or config.password
        self.token = token or config.token
        self.branch = branch

        self._api = RepoAPI(f"{owner}/{name}", host=self.host, auth=self.auth)

        # For mirror uploading: store the last revision for which we uploaded
        # When the last revision changes, that means the sync has been complete and we can upload a new batch
        self._last_upload_revision: Optional[str] = None
        self._last_upload_had_changes = False
        self._uploading_to_new_branch = False

        self.current_progress: Optional[rich.progress.Progress] = None

        if self.branch is None:
            logger.debug("Branch wasn't provided. Fetching default branch...")
            self.branch = self._api.default_branch
        logger.debug(f"Set branch: {self.branch}")

    def upload(
        self,
        local_path: Union[str, IOBase],
        commit_message=DEFAULT_COMMIT_MESSAGE,
        remote_path: str = None,
        **kwargs,
    ):
        """
        The upload function is used to upload files to the repository.
        It takes a file as an argument and logs the response status code and content.


        :param local_path: Specify the file to be uploaded
        :param commit_message: Specify a commit message
        :param remote_path: Specify the path to upload the file to
        :param **kwargs: Pass in any additional parameters that are required for the upload function
        :return: None

        """
        local_path = Path(local_path).resolve()
        if local_path.is_dir():
            if remote_path is None:
                try:
                    remote_path = local_path.relative_to(Path.cwd().resolve())
                except ValueError:
                    # local_path is outside cwd, use only its basename
                    remote_path = local_path.name
            remote_path = Path(remote_path).as_posix()
            dir_to_upload = self.directory(remote_path)
            dir_to_upload.add_dir(str(local_path), commit_message=commit_message, **kwargs)
        else:
            file_to_upload = DataSet.get_file(str(local_path), remote_path)
            self.upload_files([file_to_upload], commit_message=commit_message, **kwargs)

    def upload_files(
        self,
        files: List[FileUploadStruct],
        directory_path: str = "",
        commit_message: str = DEFAULT_COMMIT_MESSAGE,
        versioning: str = "auto",
        new_branch: str = None,
        last_commit: str = None,
        force: bool = False,
    ):
        """
        The upload_files function uploads a list of files to the specified directory.

        Args:
            files: List of Tuples of (path in repo, binaryIO) of files to upload
            directory_path: Directory in repo relative to which to upload files
            commit_message: Commit message
            versioning: Which versioning system to use to upload a file.
                Possible options: git, dvc, auto (best effort guess)
            new_branch: Create a new branch with the name of the passed argument
            last_commit: Consistency argument - last revision of the files you want to upgrade.
                Exists to prevent accidental overwrites
            force (bool): Force the upload of a file even if it is already present on the server.
                Sets last_commit to be the tip of the branch
        """

        # Truncate the commit message because the max we allow is 100 symbols
        commit_message = commit_message[:100]

        data = {
            "commit_choice": "direct",
            "commit_summary": commit_message,
            "versioning": versioning,
            "last_commit": last_commit,
            "is_dvc_dir": directory_path != "" and versioning != "git",
        }

        if new_branch is not None:
            data.update(
                {
                    "commit_choice": "commit-to-new-branch",
                    "new_branch_name": new_branch,
                }
            )
        elif self._api.is_mirror:
            # If not uploading to a new branch, and we're in a mirror - wait for the sync to complete
            self._poll_mirror_up_to_date()

        # Unset the new branch upload flag if it was set
        # (do that only after mirror poll so commit 2 on a new branch pushes correctly)
        if self._uploading_to_new_branch:
            self._uploading_to_new_branch = False

        try:
            self._last_upload_revision = self._api.last_commit_sha(self.branch)
        # New branch does not have a commit yet, so "last_upload_revision" will be None initially
        # NOTE: checking for new_branch is not enough, because it doesn't take into account
        # uploading to completely blank repos
        except BranchNotFoundError:
            self._uploading_to_new_branch = True

        if force:
            data["last_commit"] = self._last_upload_revision

        log_message(f'Uploading files ({len(files)}) to "{self._api.full_name}"...', logger)
        res = s.put(
            self.get_request_url(directory_path),
            data=data,
            files=[("files", file) for file in files],
            auth=self.auth,
            timeout=None,
        )
        self._log_upload_details(data, res, files)

        # The ETag header contains the hash of the uploaded commit,
        # check against the one we have to determine if anything changed
        if "ETag" in res.headers:
            new_tip = res.headers["ETag"]
            self._last_upload_had_changes = new_tip != self._last_upload_revision

    def _log_upload_details(self, data: Dict[str, Any], res: httpx.Response, files):
        """
        The _log_upload_details function debug logs the request URL, data, and files.
        It also prints for the user the status of their upload if it was successful
        If the response is 4xx/5xx it raises an error.

        Args:
            data: Executed request's body
            res: Server's response
            files: Uploaded file contents
        """

        logger.debug(
            f"Request URL: {res.request.url}\n"
            f"Data:\n{json.dumps(data, indent=4)}\n"
            f"Files:\n{json.dumps(list(map(str, files)), indent=4)}"
        )

        if res.status_code == HTTPStatus.OK:
            if "ETag" in res.headers:
                new_tip = res.headers["ETag"]
                if new_tip == self._last_upload_revision:
                    log_message("Upload successful, content was identical and no new commit was created", logger)
                    return
            log_message("Upload finished successfully!", logger)
        elif res.status_code == HTTPStatus.NO_CONTENT:
            log_message("Upload successful, content was identical and no new commit was created", logger)
        elif 200 < res.status_code < 300:
            log_message(f"Got unknown successful status code {res.status_code}")
        else:
            raise determine_upload_api_error(res)

    def _poll_mirror_up_to_date(self):
        """
        Synchronization lock for the mirrored repository uploading
        Since the upload is being done "through" DagsHub,
            and we rely on the change first showing up on the original repo, then being synced back to us,
            there is a possibility of uploading a new batch before the sync has been completed,
            during which the DagsHub repo is out of date with the original mirror.
        This is a client-side fix made to mitigate this.
        We poll for the change in the revision and compare it to the revision we had when we last uploaded
        When the revision changes, that means the sync has been completed and we can upload a new batch
        """
        if not self._api.is_mirror:
            return

        # Initial state - assume we can upload
        # Also can upload if last upload didn't have any changes
        if not self._uploading_to_new_branch and (
            self._last_upload_revision is None or not self._last_upload_had_changes):
            return

        poll_interval = 1.0  # seconds
        poll_timeout = 600.0
        start_time = time.time()

        if self.current_progress is not None:
            task = self.current_progress.add_task("Waiting for the mirror to sync", total=None)

            def finish():
                self.current_progress.remove_task(task)
        else:
            status = rich.status.Status("Waiting for the mirror to sync", console=rich_console)
            status.start()

            def finish():
                status.stop()

        while time.time() - start_time < poll_timeout:
            try:
                new_revision = self._api.last_commit_sha(self.branch)
            except BranchNotFoundError:
                # New branch that didn't get back to DagsHub yet
                time.sleep(poll_interval)
                continue
            if new_revision == self._last_upload_revision:
                time.sleep(poll_interval)
            else:
                finish()
                return

        finish()
        logger.warning(f"Timed out while polling for a mirror sync finishing after {poll_timeout} s. "
                       f"Trying to push anyway, which might not work.")

    @property
    def auth(self):
        """
        The auth function is used to authenticate the user with the dagshub API.
            It takes in a username and password, or token as arguments. If both are provided,
            it will use the username and password combination to
            get a token from dagshub's authentication server.
            Otherwise, if only a token is provided it will be used for authentication.

        :return: The HTTPAuth object

        """
        if self.username is not None and self.password is not None:
            return httpx.BasicAuth(self.username, self.password)
        token = self.token or dagshub.auth.get_token()
        return HTTPBearerAuth(token)

    def directory(self, path):
        """
        The directory function returns a DataSet object that represents the directory at the given path.


        :param path (str): Specify the directory that will contain the data.
                           This directory is the "root" of the dataset.
        :return: A dataset object that represents the directory at the given path

        """
        return DataSet(self, path)

    def get_request_url(self, directory):
        """
        The get_request_url function returns the URL for uploading a file to DagsHub.

        :param directory: the path to a directory within this repo on DagsHub.
            For example, if you have created your repo in such a
            way that it has two directories named data and models,
            then you could pass one of these strings into this function as an argument.
        :return: The url for uploading a file

        """
        return self.get_repo_url(CONTENT_UPLOAD_URL, directory)

    def get_files_ui_url(self, directory):
        """
        The get_files_ui_url function returns the URL for seeing the uploaded files on DagsHub.

        :param directory: the path to a directory within this repo on DagsHub.
            For example, if you have created your repo in such a
            way that it has two directories named data and models,
            then you could pass one of these strings into this function as an argument.
        :return: The url that you can navigate to in your browser to see the files

        """
        return self.get_repo_url(FILES_UI_URL, directory)

    def get_repo_url(self, url_format, directory):
        return urllib.parse.urljoin(
            self.host,
            url_format.format(
                owner=self.owner,
                reponame=self.name,
                branch=self.branch,
                path=urllib.parse.quote(directory, safe=""),
            ),
        )


class DataSet:
    def __init__(self, repo: Repo, directory: str):
        """

        :param repo (Repo object): Pass a repo object
        :param directory (str): Specify the directory of the repository
        :return: A DataSet object

        """

        self.files: Dict[os.PathLike, Tuple[os.PathLike, BinaryIO]] = {}
        self.repo = repo
        self.directory = self._clean_directory_name(directory)

    def add(self, file: Union[str, IOBase], path=None):
        """
        The add function adds a file to the list of files that will be uploaded.


        :param file (str): Specify the file to be uploaded
        :param path (str): Specify the path to upload the file
        :return: None

        """

        path, file = self.get_file(file, path)
        if file is not None:
            if path in self.files:
                log_message(
                    f'File already staged for upload on path "{path}". Overwriting', logger
                )
            self.files[path] = (path, file)

    def add_dir(self, local_path, glob_exclude="", commit_message=None, **upload_kwargs):
        """
        The add_dir function adds an entire dvc directory to a DagsHub repository.
        It does this by iterating through all the files in the given directory and uploading them one-by-one.
        The function also commits all of these changes at once, so as not to overload the API with requests.


        :param local_path  (str): Specify the local path where the dataset to upload is located
        :param glob_exclude (str): Exclude certain files from the upload process
        :param commit_message (str): Commit message
        :param upload_kwargs (dict): kwargs that are passed to the uploading function
        :return: None

        """
        upload_file_number = 100
        file_counter = 0

        progress = rich.progress.Progress(rich.progress.SpinnerColumn(), *rich.progress.Progress.get_default_columns(),
                                          rich.progress.MofNCompleteColumn(),
                                          console=rich_console, transient=True, disable=config.quiet)
        total_task = progress.add_task("Uploading files...", total=None)
        self.repo.current_progress = progress

        # If user hasn't specified versioning, then assume we're uploading dvc (this makes most sense for folders)
        if "versioning" not in upload_kwargs:
            upload_kwargs["versioning"] = "dvc"

        try:
            with progress:
                for root, dirs, files in os.walk(local_path):
                    folder_task = progress.add_task(f"Uploading files from {root}", total=len(files))

                    if commit_message is None:
                        commit_message = upload_kwargs.get("commit_message", f"Commit data points in folder {root}")
                    if "commit_message" in upload_kwargs:
                        del upload_kwargs["commit_message"]

                    if len(files) > 0:
                        for filename in files:
                            rel_file_path = posixpath.join(root, filename)
                            rel_remote_file_path = rel_file_path.replace(local_path, "")
                            if (
                                glob_exclude == ""
                                or fnmatch.fnmatch(rel_file_path, glob_exclude) is False
                            ):
                                self.add(file=rel_file_path, path=rel_remote_file_path)
                                if len(self.files) >= upload_file_number:
                                    file_counter += len(self.files)
                                    self.commit(commit_message, **upload_kwargs)
                                    progress.update(folder_task, advance=len(self.files), refresh=True)
                                    progress.update(total_task, completed=file_counter, refresh=True)
                        if len(self.files) >= upload_file_number:
                            file_counter += len(self.files)
                            self.commit(commit_message, **upload_kwargs)
                            progress.update(folder_task, advance=len(self.files), refresh=True)
                            progress.update(total_task, completed=file_counter, refresh=True)
                    progress.remove_task(folder_task)

                if len(self.files) > 0:
                    file_counter += len(self.files)
                    self.commit(commit_message, **upload_kwargs)
                    progress.update(total_task, completed=file_counter)

            log_message(f"Directory upload complete, uploaded {file_counter} files"
                        f" to {self.repo.get_files_ui_url(self.directory)}", logger)
        finally:
            self.repo.current_progress = None

    @staticmethod
    def _clean_directory_name(directory: str):
        """
        The _clean_directory_name function takes a directory name as an argument
        and returns the normalized path of that directory.
        For example, if the input is ../../../ then it will return /. If the input is ./ then it will return ..
        If there are any other characters in the string,
        they will be ignored and only alphanumeric characters (a-zA-Z0-9)
        will be kept.

        :param directory (str): Specify the directory that will be cleaned
        :return: The normalized path of the directory
                 (The directory name with the path separator normalized to a forward slash)

        """

        return posixpath.normpath(directory)

    @staticmethod
    def get_file(file: Union[str, IOBase], path: os.PathLike = None) -> FileUploadStruct:
        """
        The get_file function is a helper function that takes in either a string or an IOBase object and returns
        a tuple containing the file's name and the file itself. If no path is provided, it will default to the name of
        the file.

        :param file (Union[str, IOBase]): File to upload
        :param path (str): Desired path of the file in the repository
        :return: A tuple of the path and a file object

        """

        try:
            # if path is not provided, fall back to the file name
            if path is None:
                try:
                    path = posixpath.basename(
                        posixpath.normpath(file if type(file) is str else file.name)
                    )
                except Exception:
                    raise RuntimeError(
                        "Could not interpret your file's name. Please specify it in the keyword parameter 'path'."
                    )

            if type(file) is str:
                try:
                    f = open(file, "rb")
                    return path, f
                except IsADirectoryError:
                    raise IsADirectoryError(
                        "'file' must describe a file, not a directory."
                    )

            return path, file

        except Exception as e:
            logger.error(e)
            raise

    def _reset_dataset(self):
        """
        The _reset_dataset function clears the files attribute of a Dataset object.

        :return: None
        """
        for f in self.files.values():
            try:
                f[1].close()
            except Exception as e:
                logger.warning(f"Error closing file {f[0]}: {e}")
        self.files.clear()

    def commit(self, commit_message=DEFAULT_COMMIT_MESSAGE, *args, **kwargs):
        """
        The commit function is used to commit the files in the dataset.
        It takes a commit message as an argument,
            if no argument is passed then it return default commit message "Upload files using DagsHub client".
        The function returns nothing.


        :param commit_message (str): Set the commit message
        :param *args: Pass a non-keyworded, variable-length argument list to the function
        :param **kwargs: Pass additional parameters to the function
        :return: None
        """

        file_list = list(self.files.values())
        self.repo.upload_files(
            file_list, self.directory, commit_message=commit_message, *args, **kwargs
        )
        self._reset_dataset()



import httpx

from dagshub.upload.errors import determine_upload_api_error, UpdateNotAllowedError


def test_determine_error():
    resp = httpx.Response(400, content='{"error": "missing last_commit", "details": "file exist"}')
    err = determine_upload_api_error(resp)
    assert isinstance(err, UpdateNotAllowedError)



import types
import logging
import importlib
from urllib.parse import urljoin, quote

logger = logging.getLogger(__name__)


def multi_urljoin(*parts):
    """Shoutout to https://stackoverflow.com/a/55722792"""
    return urljoin(parts[0] + "/", "/".join(quote(part.strip("/"), safe="/") for part in parts[1:]))


def lazy_load(module_name, source_package=None, callback=None):
    if source_package is None:
        # TODO: need to have a map for commonly used imports here. Also handle dots
        source_package = module_name
    return LazyModule(module_name, source_package, callback)


class LazyModule(types.ModuleType):
    """Proxy module that lazily imports the underlying module the first time it
    is actually used.

    Shoutout to voxel51 for this :)

    Args:
        module_name: the fully-qualified module name to import
        callback (None): a callback function to call before importing the
            module
    """

    def __init__(self, module_name, source_package, callback=None):
        super().__init__(module_name)
        self._module = None
        self._callback = callback
        self._source_package = source_package

    def __getattr__(self, item):
        if self._module is None:
            self._import_module()

        return getattr(self._module, item)

    def __dir__(self):
        if self._module is None:
            self._import_module()

        return dir(self._module)

    def _import_module(self):
        # Execute callback, if any
        if self._callback is not None:
            self._callback()

        # Actually import the module
        try:
            module = importlib.import_module(self.__name__)
            self._module = module
        except ModuleNotFoundError:
            logger.warning(f"Could not import module {self.__name__}. Make sure to pip install {self._source_package}")
            raise

        # Update this object's dict so that attribute references are efficient
        # (__getattr__ is only called on lookups that fail)
        self.__dict__.update(module.__dict__)



import datetime
import json
import logging
import tempfile
from pathlib import PosixPath
from socket import gethostname, gethostbyname

import httpx
import urllib.parse
from IPython import get_ipython

from dagshub.common.helpers import log_message
from dagshub.upload import Repo

logger = logging.getLogger(__name__)


def _inside_notebook():
    return get_ipython() is not None


def _inside_colab():
    try:
        if get_ipython() and 'google.colab' in get_ipython().extension_manager.loaded:
            return True
    except Exception:
        pass
    return False


def _default_notebook_name():
    if _inside_colab():
        # Shout out to https://stackoverflow.com/a/61906730
        ip = gethostbyname(gethostname())
        filename = httpx.get(f"http://{ip}:9000/api/sessions").json()[0]["name"]
        filename = urllib.parse.unquote(filename)
        if not filename.endswith(".ipynb"):
            filename = filename + ".ipynb"
        return filename
    return f"notebook-{datetime.datetime.utcnow().strftime('%Y-%m-%d')}.ipynb"


def save_notebook(repo, path="", branch=None, commit_message=None, versioning='git') -> None:
    """
    IPython wrapper for saving notebooks.

    :param path (str): Where to save the notebook within the repository (including the filename).
        If filename is not specified, we'll save it as "notebook-{datetime.now}.ipynb" under specified folder
    :prama repo (str): repository in the format of "user/repo"
    :param branch (str): The branch under which the notebook should be saved.
        Will commit to the default repo branch if not specified
    :param commit_message (str): The commit message for the update
    :param versioning (str): ['git'|'dvc'] The VCS used to version the notebook
    """

    if not _inside_notebook():
        log_message('Trying to save a notebook while not being in an IPython environment. No notebook will be saved',
                    logger)
        return

    # Handle file path
    file_path = PosixPath(path)
    if file_path.name != "." and "." not in file_path.name:
        file_path /= _default_notebook_name()
    file_path = "/" / file_path

    # Handle commit message
    if commit_message is None:
        commit_message = f"Uploaded notebook {file_path.name}"

    # Handle repo name
    parsed_repo = repo.split("/")
    if len(parsed_repo) != 2:
        raise RuntimeError(f'Repo format has to be "user/repo" (got {repo})')
    owner, repo = parsed_repo

    # Upload notebook
    with tempfile.TemporaryDirectory() as tmp:
        out_path = f"{tmp}/{file_path.name}"
        if _inside_colab():
            from google.colab import _message  # If inside colab, this import is guaranteed
            with open(out_path, 'w') as file:
                file.write(json.dumps(_message.blocking_request('get_ipynb')["ipynb"], indent=4))
        else:
            get_ipython().run_line_magic('notebook', out_path)

        repo = Repo(owner, repo, branch=branch)
        repo.upload(out_path,
                    remote_path=file_path.as_posix(),
                    commit_message=commit_message,
                    versioning=versioning,
                    force=True)



import random
from multiprocessing import Process
from typing import TYPE_CHECKING, Union

from dagshub.common.util import lazy_load
from dagshub.data_engine.client.loaders.base import DagsHubDataset

np = lazy_load("numpy")
tf = lazy_load("tensorflow")

if TYPE_CHECKING:
    import tensorflow as tf


class TensorFlowDataset(DagsHubDataset):
    def __init__(self, *args, **kwargs):
        self.tensorlib = Tensorizers
        super().__init__(*args, **kwargs)
        self.signature = tuple(
            tf.TensorSpec.from_tensor(tensor) for tensor in next(self.generator())
        )

    def generator(self):
        for idx in range(len(self)):
            yield self[idx]


class TensorFlowDataLoader(tf.keras.utils.Sequence):
    def __init__(self, dataset, batch_size=1, shuffle=True, seed=None):
        self.dataset = dataset
        self.batch_size = batch_size
        self.shuffle = shuffle

        if seed:
            random.seed(seed)
            np.random.seed(seed)

        self.indices = {}
        self.on_epoch_end()

        self.dataset.builder.order = self.indices
        if self.dataset.builder.strategy == "background":
            Process(target=self.dataset.builder.pull).start()

    def __len__(self) -> int:
        return self.dataset.__len__() // self.batch_size

    def __getitem__(self, index: int) -> tf.Tensor:
        samples = [
            self.dataset.__getitem__(index)
            for index in self.indices[
                index * self.batch_size : (index + 1) * self.batch_size
            ]
        ]
        batch = [
            [] for _ in range(len(samples[0]))
        ]  # [[]] * len(samples[0]) creates lists shallowly
        for sample in samples:
            for idx, tensor in enumerate(sample):
                batch[idx].append(tensor)

        return [tf.stack(column) for column in batch]

    def on_epoch_end(self) -> None:
        self.indices = np.arange(self.dataset.__len__())
        if self.shuffle:
            np.random.shuffle(self.indices)


class Tensorizers:
    @staticmethod
    def image(filepath: str) -> tf.Tensor:
        return tf.convert_to_tensor(tf.keras.utils.load_img(filepath))

    @staticmethod
    def audio(filepath: str) -> tf.Tensor:
        raise NotImplementedError("Coming Soon!")

    @staticmethod
    def video(filepath: str) -> tf.Tensor:
        raise NotImplementedError("Coming Soon!")

    @staticmethod
    def numeric(num: Union[float, int]) -> tf.Tensor:
        return tf.convert_to_tensor(num)



import enum
import logging
from typing import TYPE_CHECKING, Optional, Union, Dict, Type

from treelib import Tree, Node

from dagshub.data_engine.client.models import MetadataFieldType
from dagshub.data_engine.model.errors import WrongOperatorError

if TYPE_CHECKING:
    from dagshub.data_engine.model.datasource import Datasource

logger = logging.getLogger(__name__)

_metadataTypeLookup = {
    int: MetadataFieldType.INTEGER,
    bool: MetadataFieldType.BOOLEAN,
    float: MetadataFieldType.FLOAT,
    str: MetadataFieldType.STRING,
    bytes: MetadataFieldType.BLOB,
}

_metadataTypeCustomConverters = {
    bool: lambda x: x.lower() == "true",
}

_metadataTypeLookupReverse: Dict[str, Type] = {}
for k, v in _metadataTypeLookup.items():
    _metadataTypeLookupReverse[v.value] = k


class FieldFilterOperand(enum.Enum):
    EQUAL = "EQUAL"
    GREATER_THAN = "GREATER_THAN"
    GREATER_EQUAL_THAN = "GREATER_EQUAL_THAN"
    LESS_THAN = "LESS_THAN"
    LESS_EQUAL_THAN = "LESS_EQUAL_THAN"
    CONTAINS = "CONTAINS"
    IS_NULL = "IS_NULL"


fieldFilterOperandMap = {
    "eq": FieldFilterOperand.EQUAL,
    "gt": FieldFilterOperand.GREATER_THAN,
    "ge": FieldFilterOperand.GREATER_EQUAL_THAN,
    "lt": FieldFilterOperand.LESS_THAN,
    "le": FieldFilterOperand.LESS_EQUAL_THAN,
    "contains": FieldFilterOperand.CONTAINS,
    "isnull": FieldFilterOperand.IS_NULL,
}

fieldFilterOperandMapReverseMap: Dict[str, str] = {}
for k, v in fieldFilterOperandMap.items():
    fieldFilterOperandMapReverseMap[v.value] = k


class DatasourceQuery:
    def __init__(self, column_or_query: Optional[Union[str, "DatasourceQuery"]] = None):
        self._operand_tree: Tree = Tree()
        self._column_filter: Optional[str] = None  # for storing filters when user does ds["column"]
        if type(column_or_query) is str:
            # If it's ds["column"] then the root node is just the column name
            self._column_filter = column_or_query
        elif column_or_query is not None:
            self._operand_tree.create_node(column_or_query)

    def __repr__(self):
        if self.is_empty:
            return "Query: empty"
        return f"Query: {self.to_dict()}"

    @property
    def column_filter(self) -> Optional[str]:
        return self._column_filter

    def compose(self, op: str, other: Optional[Union[str, int, float, "DatasourceQuery", "Datasource"]]):
        if self._column_filter is not None:
            # Just the column is in the query - compose into a tree
            self._operand_tree.create_node(op, data={"field": self._column_filter, "value": other})
            self._column_filter = None
        elif op == "isnull":
            # Can only do isnull on the column filter, if we got here, there's something wrong
            raise RuntimeError("is_null operation can only be done on a column (e.g. ds['col1'].is_null())")
        elif op == "not":
            new_tree = Tree()
            not_node = new_tree.create_node("not")
            new_tree.paste(not_node.identifier, self._operand_tree)
            self._operand_tree = new_tree
        else:
            # The query is an actual query with a tree - make a subtree
            if type(other) is not DatasourceQuery:
                raise RuntimeError(f"Expected other argument to be a dataset, got {type(other)} instead")
            if op not in ["and", "or"]:
                raise RuntimeError(f"Cannot use operator '{op}' to chain two queries together.\r\n"
                                   f"Queries:\r\n"
                                   f"\t{self}\r\n"
                                   f"\t{other}\r\n")
            # Don't compose with an empty query, carry the other instead
            if self.is_empty:
                self._operand_tree = other._operand_tree
                return
            elif other.is_empty:
                return
            composite_tree = Tree()
            root_node = composite_tree.create_node(op)
            composite_tree.paste(root_node.identifier, self._operand_tree)
            composite_tree.paste(root_node.identifier, other._operand_tree)
            self._operand_tree = composite_tree

    @property
    def _operand_root(self) -> Node:
        return self._operand_tree[self._operand_tree.root]

    def serialize_graphql(self):
        if self.is_empty:
            return None
        return self._serialize_node(self._operand_root, self._operand_tree)

    @staticmethod
    def _serialize_node(node: Node, tree: Tree) -> Dict:
        operand = node.tag
        if operand in ["and", "or"]:
            # recursively serialize children subqueries
            return {operand: [DatasourceQuery._serialize_node(child, tree) for child in tree.children(node.identifier)]}
        if operand == "not":
            assert len(tree.children(node.identifier)) == 1
            child = tree.children(node.identifier)[0]
            serialized = DatasourceQuery._serialize_node(child, tree)
            serialized["not"] = True
            return serialized
        else:
            query_op = fieldFilterOperandMap.get(operand)
            if query_op is None:
                raise WrongOperatorError(f"Operator {operand} is not supported")
            key = node.data["field"]
            value = node.data["value"]
            value_type = _metadataTypeLookup[type(value)].value
            if value_type is None:
                raise RuntimeError(f"Value type {value_type} is not supported for querying.\r\n"
                                   f"Supported types: {list(_metadataTypeLookup.keys())}")
            return {
                "filter": {
                    "key": key,
                    "value": str(value),
                    "valueType": value_type,
                    "comparator": query_op.value,
                }
            }

    @staticmethod
    def deserialize(serialized_query: Dict) -> "DatasourceQuery":
        q = DatasourceQuery()
        op_tree = Tree()

        DatasourceQuery._deserialize_node(serialized_query, op_tree)

        q._operand_tree = op_tree
        return q

    @staticmethod
    def _deserialize_node(node_dict: Dict, tree: Tree, parent_node=None) -> None:
        keys = list(node_dict.keys())

        is_negative = node_dict.get("not", False)
        if is_negative:
            # If operation is negative - prepend a "not" node to the node we'll be adding
            neg_node = Node(tag="not")
            tree.add_node(neg_node, parent_node)
            parent_node = neg_node

        op_type = keys[0]
        val = node_dict[op_type]
        # Types: and, or, filter
        if op_type == "filter":
            comparator = fieldFilterOperandMapReverseMap[val["comparator"]]
            key = val["key"]
            value_type = _metadataTypeLookupReverse[val["valueType"]]
            converter = _metadataTypeCustomConverters.get(value_type, lambda x: value_type(x))
            value = converter(val["value"])
            node = Node(tag=comparator, data={"field": key, "value": value})
            tree.add_node(node, parent_node)
        elif op_type in ("and", "or"):
            main_node = Node(tag=op_type)
            tree.add_node(main_node, parent_node)
            for nested_node in val:
                DatasourceQuery._deserialize_node(nested_node, tree, main_node)
        else:
            raise RuntimeError(f"Unknown serialized query dict: {node_dict}")

    def to_dict(self):
        return self._operand_tree.to_dict(with_data=True)

    def __deepcopy__(self, memodict={}):
        q = DatasourceQuery()
        if self._column_filter is not None:
            q._column_filter = self._column_filter
        else:
            q._operand_tree = Tree(tree=self._operand_tree, deep=True)
        return q

    @property
    def is_empty(self):
        return self._column_filter is not None or self._operand_tree.root is None



import logging
import re
from dataclasses import dataclass, field
from pathlib import PurePosixPath
from typing import Optional, Union, Mapping, Any, Dict, List

from dagshub.common.api.repo import RepoAPI
from dagshub.data_engine.client.data_client import DataClient
from dagshub.data_engine.client.models import DatasourceType, DatasourceResult, PreprocessingStatus, \
    MetadataFieldSchema
from dagshub.data_engine.model.datapoint import Datapoint
from dagshub.data_engine.model.errors import DatasourceAlreadyExistsError, DatasourceNotFoundError

try:
    from functools import cached_property
except ImportError:
    from cached_property import cached_property

logger = logging.getLogger(__name__)

path_regexes = {
    DatasourceType.BUCKET: re.compile(r"(?P<schema>s3|gs)://(?P<bucket>[\w\-._]+)(?P<prefix>/.*)?"),
    DatasourceType.REPOSITORY: re.compile(r"repo://(?P<user>[\w\-_.]+)/(?P<repo>[\w\-_.]+)(?P<prefix>/.*)?"),
}

expected_formats = {
    DatasourceType.BUCKET: "s3|gs://bucket-name/prefix",
    DatasourceType.REPOSITORY: "repo://owner/reponame/prefix",
}


class InvalidPathFormatError(Exception):
    pass


@dataclass
class DatasourceState:
    repo: str
    name: Optional[str] = field(default=None)
    id: Optional[Union[int, str]] = field(default=None)

    source_type: DatasourceType = field(init=False)
    preprocessing_status: PreprocessingStatus = field(init=False)
    path: str = field(init=False)
    client: DataClient = field(init=False)
    repoApi: RepoAPI = field(init=False)
    metadata_fields: List[MetadataFieldSchema] = field(init=False)

    _revision: Optional[str] = field(init=False, default=None)

    def __post_init__(self):
        self.client = DataClient(self.repo)
        self.repoApi = RepoAPI(self.repo)
        if hasattr(self, "source_type") and self.source_type == DatasourceType.REPOSITORY:
            self.revision = self.path_parts()["revision"]
        if not hasattr(self, "metadata_fields"):
            self.metadata_fields = []

    @property
    def revision(self) -> str:
        """Used for repository sources, provides branch/revision from which to download files"""
        if self._revision is None:
            logger.warning("Revision wasn't set, assuming default repo branch")
            self.revision = self.repoApi.default_branch
        return self._revision

    @revision.setter
    def revision(self, val: str):
        self._revision = val

    def create(self):
        # Check that a datasource with this name doesn't exist
        datasources = self.client.get_datasources(self.id, self.name)
        if len(datasources) != 0:
            raise DatasourceAlreadyExistsError(self)

        create_result = self.client.create_datasource(self)
        self._update_from_ds_result(create_result)

    def get_from_dagshub(self):
        sources = self.client.get_datasources(self.id, self.name)
        if len(sources) == 0:
            raise DatasourceNotFoundError(self)
        elif len(sources) > 1:
            raise RuntimeError(
                f"Got too many ({len(sources)}) datasources with name '{self.name}' or id. Something went wrong")
        self._update_from_ds_result(sources[0])

    def content_path(self, path: Union[str, Datapoint, Mapping[str, Any]]) -> str:
        """
        Returns the url for the content path of a specified path
        """
        path = self._extract_path(path).strip("/")
        return self.root_content_path + "/" + path

    def raw_path(self, path: Union[str, Datapoint, Mapping[str, Any]]) -> str:
        """
        Returns the url for the download path of a specified path
        """
        path = self._extract_path(path).strip("/")
        return self.root_raw_path + "/" + path

    @property
    def source_prefix(self) -> PurePosixPath:
        parts = self.path_parts()
        if "prefix" in parts:
            return PurePosixPath(parts["prefix"].strip("/"))
        else:
            return PurePosixPath()

    def file_path(self, path: Union[str, Datapoint, Mapping[str, Any]]) -> PurePosixPath:
        """
        Returns the generic path of the path in the repo (adds prefix)
        """
        return self.source_prefix / self._extract_path(path)

    def blob_path(self, sha: str) -> str:
        """
        Returns the path for the blob of a datasource
        """
        return f"{self.repoApi.data_engine_url}/blob/{sha}"

    @cached_property
    def root_content_path(self) -> str:
        """
        Returns the root content path of the dataset for listing folders
        This is just a "prefix" of the datasource relative to the repo.
        In order to build a path of an entity you need to concatenate the path to this root
        """
        return self._root_path("content")

    @cached_property
    def root_raw_path(self):
        """
        Returns the root raw path of the dataset for downloading files
        This is just a "prefix" of the datasource relative to the repo.
        In order to build a path of an entity you need to concatenate the path to this root
        """
        return self._root_path("raw")

    def _root_path(self, path_type):
        assert path_type in ["raw", "content"]
        parts = self.path_parts()
        if self.source_type == DatasourceType.BUCKET:
            path_elems = [parts["schema"], parts["bucket"]]
            if parts["prefix"] is not None:
                path_elems.append(parts["prefix"])
            path_prefix = "/".join(path_elems)
            if path_type == "raw":
                return self.repoApi.storage_raw_api_url(path_prefix)
            elif path_type == "content":
                return self.repoApi.storage_content_api_url(path_prefix)
        elif self.source_type == DatasourceType.REPOSITORY:
            prefix = parts["prefix"]
            if prefix is None:
                prefix = ""
            # Assuming repo://user/repo is always the same user/repo we work with
            if path_type == "raw":
                return self.repoApi.raw_api_url(prefix, self.revision)
            elif path_type == "content":
                return self.repoApi.content_api_url(prefix, self.revision)
        elif self.source_type == DatasourceType.CUSTOM:
            raise NotImplementedError
        raise NotImplementedError

    def path_parts(self) -> Dict[str, str]:
        """
        Validates the provided path + returns a dictionary with elements that might be relevant for constructing paths
        """
        regex = path_regexes[self.source_type]
        match = regex.fullmatch(self.path)
        if match is None:
            raise InvalidPathFormatError(f"{self.path} is not valid path format for type {self.source_type}.\n"
                                         f"Expected format: {expected_formats[self.source_type]}")
        res = match.groupdict()
        # For repository type - handle revision that is in format of repo://user/repo/branch:prefix
        # Couldn't do that with regexes, so handling it here
        if self.source_type == DatasourceType.REPOSITORY:
            res["revision"] = None
            if res["prefix"] is not None:
                prefix = res["prefix"]
                if ":" in prefix:
                    revision, prefix = prefix.split(":", 1)
                    res["revision"] = revision.strip("/")
                    if prefix.isspace():
                        prefix = None
                    elif not prefix.startswith("/"):
                        prefix = "/" + prefix
                    res["prefix"] = prefix
        return res

    @staticmethod
    def _extract_path(val: Union[str, Datapoint, Mapping[str, Any]]) -> str:
        if type(val) is str:
            return val
        elif type(val) is Datapoint:
            return val.path
        return val["path"]

    def _update_from_ds_result(self, ds: DatasourceResult):
        self.id = ds.id
        self.name = ds.name
        self.path = ds.rootUrl
        self.source_type = ds.type
        self.preprocessing_status = ds.preprocessingStatus
        self.metadata_fields = [] if ds.metadataFields is None else ds.metadataFields
        if self.source_type == DatasourceType.REPOSITORY:
            self.revision = self.path_parts()["revision"]

    @staticmethod
    def from_gql_result(repo: str, res: DatasourceResult):
        ds = DatasourceState(repo)
        ds._update_from_ds_result(res)
        return ds



from dataclasses import dataclass
from pathlib import PurePosixPath
from typing import Optional, Dict, List

try:
    from functools import cached_property
except ImportError:
    from cached_property import cached_property


@dataclass
class RepoAPIResponse:
    id: int
    owner: "UserAPIResponse"
    name: str
    full_name: str
    description: str
    private: bool
    fork: bool
    parent: Optional["RepoAPIResponse"]
    empty: bool
    mirror: bool
    size: int
    html_url: str
    ssh_url: Optional[str]
    clone_url: str
    website: Optional[str]
    stars_count: int
    forks_count: int
    watchers_count: int
    open_issues_count: int
    default_branch: str
    created_at: str
    updated_at: str
    permissions: Optional[Dict[str, bool]]


@dataclass
class UserAPIResponse:
    id: int
    login: str
    full_name: str
    avatar_url: Optional[str]
    public_email: Optional[str]
    website: Optional[str]
    company: Optional[str]
    description: Optional[str]
    username: str


@dataclass
class BranchAPIResponse:
    name: str
    commit: "CommitAPIResponse"


@dataclass
class CommitAPIResponse:
    id: str
    message: str
    url: str
    author: Optional["GitUser"]
    committer: Optional["GitUser"]
    added: Optional[List[str]]
    removed: Optional[List[str]]
    modified: Optional[List[str]]
    timestamp: str


@dataclass
class GitUser:
    name: str
    email: str
    username: str


@dataclass
class StorageAPIEntry:
    name: str
    protocol: str
    list_path: str

    @cached_property
    def full_path(self):
        return f"{self.protocol}/{self.name}"

    @cached_property
    def path_in_mount(self) -> PurePosixPath:
        return PurePosixPath(".dagshub/storage") / self.protocol / self.name


@dataclass
class ContentAPIEntry:
    path: str
    # Possible values: dir, file, storage
    type: str
    size: int
    hash: str
    # Possible values: git, dvc, bucket
    versioning: str
    download_url: str
    content_url: Optional[str]  # TODO: remove Optional once content_url is exposed in API


@dataclass
class StorageContentAPIResult:
    entries: List[ContentAPIEntry]
    next_token: Optional[str]



__version__ = "0.3.2_1"
from .logger import DAGsHubLogger, dagshub_logger
from .common.init import init
from .upload.wrapper import upload_files

__all__ = [
    DAGsHubLogger,
    dagshub_logger,
    init,
    upload_files
]



import appdirs
import os
from urllib.parse import urlparse
from dagshub import __version__
from httpx._client import USER_AGENT

HOST_KEY = "DAGSHUB_CLIENT_HOST"
DEFAULT_HOST = "https://dagshub.com"
CLIENT_ID_KEY = "DAGSHUB_CLIENT_ID"
DEFAULT_CLIENT_ID = "32b60ba385aa7cecf24046d8195a71c07dd345d9657977863b52e7748e0f0f28"
TOKENS_CACHE_LOCATION_KEY = "DAGSHUB_CLIENT_TOKENS_CACHE"
DEFAULT_TOKENS_CACHE_LOCATION = os.path.join(
    appdirs.user_cache_dir("dagshub"), "tokens"
)
TOKENS_CACHE_SCHEMA_VERSION = "1"
DAGSHUB_USER_TOKEN_KEY = "DAGSHUB_USER_TOKEN"
DAGSHUB_USERNAME_KEY = "DAGSHUB_USERNAME"
DAGSHUB_PASSWORD_KEY = "DAGSHUB_PASSWORD"
HTTP_TIMEOUT_KEY = "DAGSHUB_HTTP_TIMEOUT"
DAGSHUB_QUIET_KEY = "DAGSHUB_QUIET"

parsed_host = urlparse(os.environ.get(HOST_KEY, DEFAULT_HOST))
hostname = parsed_host.hostname
host = parsed_host.geturl().rstrip("/")
client_id = os.environ.get(CLIENT_ID_KEY, DEFAULT_CLIENT_ID)
cache_location = os.environ.get(
    TOKENS_CACHE_LOCATION_KEY, DEFAULT_TOKENS_CACHE_LOCATION
)
token = os.environ.get(DAGSHUB_USER_TOKEN_KEY)
username = os.environ.get(DAGSHUB_USERNAME_KEY)
password = os.environ.get(DAGSHUB_PASSWORD_KEY)
custom_user_agent_suffix = f" dagshub-client-python/{__version__}"
requests_headers = {"user-agent": USER_AGENT + custom_user_agent_suffix}
http_timeout = os.environ.get(HTTP_TIMEOUT_KEY, 30)
REPO_INFO_URL = "api/v1/repos/{owner}/{reponame}"

quiet = bool(os.environ.get(DAGSHUB_QUIET_KEY, False))

# DVC config templates
CONFIG_GITIGNORE = "/config.local\n/tmp\n/cache"

RECOMMENDED_ANNOTATE_LIMIT_KEY = "RECOMMENDED_ANNOTATE_LIMIT"
recommended_annotate_limit = os.environ.get(RECOMMENDED_ANNOTATE_LIMIT_KEY, 1e5)

DATAENGINE_METADATA_UPLOAD_BATCH_SIZE_KEY = "DAGSHUB_DE_METADATA_UPLOAD_BATCH_SIZE"
dataengine_metadata_upload_batch_size = os.environ.get(DATAENGINE_METADATA_UPLOAD_BATCH_SIZE_KEY, 15000)

DISABLE_ANALYTICS_KEY = "DAGSHUB_DISABLE_ANALYTICS"
disable_analytics = "DAGSHUB_DISABLE_ANALYTICS" in os.environ



from pathlib import Path


class FilesystemAlreadyMountedError(Exception):
    def __init__(self, path: Path, repo: str, revision: str):
        self.path = path
        self.repo = repo
        self.revision = revision

    def __str__(self):
        return f"There is already a filesystem mounted at path {self.path.absolute()} " \
               f"({self.repo} revision {self.revision})" \
               f"\ndel() the filesystem object in use if you want to switch the mounted filesystem"



import hashlib
from typing import Optional, Dict
import logging
import urllib
import uuid
import httpx
import webbrowser

from dagshub.common import config, rich_console

logger = logging.getLogger(__name__)


def oauth_flow(
    host: str,
    client_id: Optional[str] = None
) -> Dict:

    host = host.strip("/")
    dagshub_url = urllib.parse.urljoin(host, "login/oauth")
    client_id = client_id or config.client_id
    state = uuid.uuid4()
    middle_man_request_id = hashlib.sha256(uuid.uuid4().bytes).hexdigest()
    auth_link = f"{dagshub_url}/authorize?state={state}&client_id={client_id}" \
                f"&middleman_request_id={middle_man_request_id}"

    webbrowser.open(auth_link)

    rich_console.print("[bold]:exclamation::exclamation::exclamation: AUTHORIZATION REQUIRED "
                       ":exclamation::exclamation::exclamation:[/bold]", justify="center")
    # Doing raw prints here because the rich syntax breaks in colab
    # Printing them line by line, because the link has to be its own print in order for Colab parser to correctly parse
    # the whole link
    print("\n\nOpen the following link in your browser to authorize the client:")
    print(auth_link)
    print("\n")

    with rich_console.status("Waiting for authorization"):
        res = httpx.post(
            f"{dagshub_url}/middleman",
            data={"request_id": middle_man_request_id}, timeout=None
        )

    if res.status_code != 200:
        raise Exception(
            f"Error while getting OAuth code: HTTP {res.status_code}, Body: {res.json()}"
        )
    code = res.json()

    res = httpx.post(
        f"{dagshub_url}/access_token",
        data={"client_id": client_id, "code": code, "state": str(state)}
    )

    if res.status_code != 200:
        raise Exception(
            f"Error while getting OAuth token: HTTP {res.status_code}, Body: {res.json()}"
        )
    token = res.json()

    logger.debug(f"Got token: {token}")
    return token


class OauthNonInteractiveShellException(Exception):
    pass



import base64
import gzip
import json
import logging
import math
import os.path
import time
import webbrowser
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union, Set, ContextManager

import rich.progress
from dataclasses_json import dataclass_json, config
from pathvalidate import sanitize_filepath

import dagshub.common.config
from dagshub.common import rich_console
from dagshub.common.analytics import send_analytics_event
from dagshub.common.helpers import prompt_user, http_request, log_message
from dagshub.common.rich_util import get_rich_progress
from dagshub.common.util import lazy_load, multi_urljoin
from dagshub.data_engine.client.models import PreprocessingStatus, MetadataFieldType, MetadataFieldSchema, \
    autogenerated_columns
from dagshub.data_engine.model.datapoint import Datapoint
from dagshub.data_engine.model.errors import WrongOperatorError, WrongOrderError, DatasetFieldComparisonError, \
    FieldNotFoundError
from dagshub.data_engine.model.query import DatasourceQuery, _metadataTypeLookup, _metadataTypeLookupReverse

if TYPE_CHECKING:
    from dagshub.data_engine.model.query_result import QueryResult
    from dagshub.data_engine.model.datasource_state import DatasourceState
    import fiftyone as fo
    import pandas
else:
    plugin_server_module = lazy_load("dagshub.data_engine.voxel_plugin_server.server")
    fo = lazy_load("fiftyone")

logger = logging.getLogger(__name__)


@dataclass_json
@dataclass
class DatapointMetadataUpdateEntry(json.JSONEncoder):
    url: str
    key: str
    value: str
    valueType: MetadataFieldType = field(
        metadata=config(
            encoder=lambda val: val.value
        )
    )
    allowMultiple: bool = False


class Datasource:

    def __init__(self, datasource: "DatasourceState", query: Optional[DatasourceQuery] = None):
        self._source = datasource
        if query is None:
            query = DatasourceQuery()
        self._query = query

        self.serialize_gql_query_input()

    @property
    def source(self) -> "DatasourceState":
        return self._source

    def clear_query(self):
        """
        This function clears the query assigned to this datasource.
        Once you clear the query, next time you try to get datapoints, you'll get all the datapoints in the datasource
        """
        self._query = DatasourceQuery()

    def __deepcopy__(self, memodict={}) -> "Datasource":
        res = Datasource(self._source, self._query.__deepcopy__())
        return res

    def get_query(self):
        return self._query

    @property
    def annotation_fields(self) -> List[str]:
        # TODO: once the annotation type is implemented, expose those columns here
        return ["annotation"]

    def serialize_gql_query_input(self):
        return {
            "query": self._query.serialize_graphql(),
        }

    def sample(self, start: Optional[int] = None, end: Optional[int] = None):
        if start is not None:
            logger.warning("Starting slices is not implemented for now")
        return self._source.client.sample(self, end, include_metadata=True)

    def head(self, size=100) -> "QueryResult":
        """
        Executes the query and returns a QueryResult object containing first <size> datapoints

        Args:
            size: how many datapoints to get. Default is 100
        """
        self._check_preprocess()
        send_analytics_event("Client_DataEngine_DisplayTopResults", repo=self.source.repoApi)
        return self._source.client.head(self, size)

    def all(self) -> "QueryResult":
        """
        Executes the query and returns a QueryResult object containing all datapoints
        """
        self._check_preprocess()
        return self._source.client.get_datapoints(self)

    def _check_preprocess(self):
        self.source.get_from_dagshub()
        if (self.source.preprocessing_status == PreprocessingStatus.IN_PROGRESS or
            self.source.preprocessing_status == PreprocessingStatus.AUTO_SCAN_IN_PROGRESS):
            logger.warning(
                f"Datasource {self.source.name} is currently in the progress of rescanning. "
                f"Values might change if you requery later")

    def metadata_context(self) -> ContextManager["MetadataContextManager"]:
        """
        Returns a metadata context, that you can upload metadata through via update_metadata
        Once the context is exited, all metadata is uploaded in one batch

        with df.metadata_context() as ctx:
            ctx.update_metadata(["file1", "file2"], {"key1": True, "key2": "value"})

        """

        # Need to have the context manager inside a wrapper to satisfy MyPy + PyCharm type hinter
        @contextmanager
        def func():
            self.source.get_from_dagshub()
            send_analytics_event("Client_DataEngine_addEnrichments", repo=self.source.repoApi)
            ctx = MetadataContextManager(self)
            yield ctx
            self._upload_metadata(ctx.get_metadata_entries())

        return func()

    def upload_metadata_from_dataframe(self, df: "pandas.DataFrame", path_column: Optional[Union[str, int]] = None):
        """
        Uploads metadata from a pandas dataframe
        path_column can either be a name of the column with the data or its index.
        This will be the column from which the datapoints are extracted.
        All the other columns are treated as metadata to upload
        If path_column is not specified, the first column is used as the datapoints
        """
        self.source.get_from_dagshub()
        send_analytics_event("Client_DataEngine_addEnrichmentsWithDataFrame", repo=self.source.repoApi)
        self._upload_metadata(self._df_to_metadata(df, path_column, multivalue_fields=self._get_multivalue_fields()))

    def _get_multivalue_fields(self) -> Set[str]:
        res = set()
        for col in self.source.metadata_fields:
            if col.multiple:
                res.add(col.name)
        return res

    def _df_to_metadata(self, df: "pandas.DataFrame", path_column: Optional[Union[str, int]] = None,
                        multivalue_fields=set()) -> List[
        DatapointMetadataUpdateEntry]:
        res: List[DatapointMetadataUpdateEntry] = []
        if path_column is None:
            path_column = df.columns[0]
        elif type(path_column) is str:
            if path_column not in df.columns:
                raise RuntimeError(f"Column {path_column} does not exist in the dataframe")
        elif type(path_column) is int:
            path_column = df.columns[path_column]

        # objects are actually mixed and not guaranteed to be string, but this should cover most use cases
        if df.dtypes[path_column] != "object":
            raise RuntimeError(f"Column {path_column} doesn't have strings")

        field_value_types = {f.name: f.valueType for f in self.fields}

        for _, row in df.iterrows():
            datapoint = row[path_column]
            for key, val in row.items():
                if key == path_column:
                    continue
                key = str(key)
                if key in autogenerated_columns:
                    continue
                if val is None:
                    continue
                # ONLY FOR PANDAS: since pandas doesn't distinguish between None and NaN, don't upload it
                if type(val) is float and math.isnan(val):
                    continue
                if type(val) is list:
                    if key not in multivalue_fields:
                        multivalue_fields.add(key)
                        # Promote all the existing uploading metadata to multivalue
                        for update_entry in res:
                            if update_entry.key == key:
                                update_entry.allowMultiple = True
                    for sub_val in val:
                        value_type = field_value_types.get(key)
                        if value_type is None:
                            value_type = _metadataTypeLookup[type(sub_val)]
                            field_value_types[key] = value_type
                        # Don't override bytes if they're not bytes - probably just undownloaded values
                        if value_type == MetadataFieldType.BLOB and type(sub_val) is not bytes:
                            continue
                        # Pandas quirk - integers are floats on the backend
                        if value_type == MetadataFieldType.INTEGER:
                            sub_val = int(sub_val)
                        if type(sub_val) is bytes:
                            sub_val = MetadataContextManager.wrap_bytes(sub_val)
                        res.append(DatapointMetadataUpdateEntry(
                            url=datapoint,
                            key=key,
                            value=str(sub_val),
                            valueType=value_type,
                            allowMultiple=True
                        ))
                else:
                    value_type = field_value_types.get(key)
                    if value_type is None:
                        value_type = _metadataTypeLookup[type(val)]
                        field_value_types[key] = value_type
                    # Don't override bytes if they're not bytes - probably just undownloaded values
                    if value_type == MetadataFieldType.BLOB and type(val) is not bytes:
                        continue
                    # Pandas quirk - integers are floats on the backend
                    if value_type == MetadataFieldType.INTEGER:
                        val = int(val)
                    if type(val) is bytes:
                        val = MetadataContextManager.wrap_bytes(val)
                    res.append(DatapointMetadataUpdateEntry(
                        url=datapoint,
                        key=key,
                        value=str(val),
                        valueType=value_type,
                        allowMultiple=key in multivalue_fields
                    ))
        return res

    def delete_source(self, force: bool = False):
        """
        Delete the record of this datasource
        This will remove ALL the datapoints + metadata associated with the datasource
        """
        prompt = f"You are about to delete datasource \"{self.source.name}\" for repo \"{self.source.repo}\"\n" \
                 f"This will remove the datasource and ALL datapoints " \
                 f"and metadata records associated with the source."
        if not force:
            user_response = prompt_user(prompt)
            if not user_response:
                print("Deletion cancelled")
                return
        self.source.client.delete_datasource(self)

    def scan_source(self):
        """
        This function fires a call to the backend to rescan the datapoints.
        Call this function whenever you uploaded new files and want them to appear when querying the datasource,
        Or if you changed existing file contents and want their metadata to be updated automatically.

        Notes about automatically scanned metadata:
        1. Only new datapoints (files) will be added.
           If files were removed from the source, their metadata will still remain,
           and they will still be returned from queries on the datasource.
           An API to actively remove metadata will be available soon.
        2. Some metadata fields will be automatically scanned and updated by DagsHub based on this scan -
           the list of automatic metadata fields is growing frequently!
        """
        logger.debug("Rescanning datasource")
        self.source.client.scan_datasource(self)

    def _upload_metadata(self, metadata_entries: List[DatapointMetadataUpdateEntry]):

        progress = get_rich_progress(rich.progress.MofNCompleteColumn())

        upload_batch_size = dagshub.common.config.dataengine_metadata_upload_batch_size
        total_entries = len(metadata_entries)
        total_task = progress.add_task(f"Uploading metadata (batch size {upload_batch_size})...",
                                       total=total_entries)

        with progress:
            for start in range(0, total_entries, upload_batch_size):
                entries = metadata_entries[start:start + upload_batch_size]
                logger.debug(f"Uploading {len(entries)} metadata entries...")
                self.source.client.update_metadata(self, entries)
                progress.update(total_task, advance=upload_batch_size)
            progress.update(total_task, completed=total_entries, refresh=True)

        # Update the status from dagshub, so we get back the new metadata columns
        self.source.get_from_dagshub()

    def save_dataset(self, name: str):
        """
        Save the dataset, which is a combination of datasource + query, on the backend.
        That way you can persist and share your queries on the backend
        You can get the dataset back by calling `datasources.get_dataset(repo, name)`
        """
        send_analytics_event("Client_DataEngine_QuerySaved", repo=self.source.repoApi)

        self.source.client.save_dataset(self, name)
        log_message(f"Dataset {name} saved")

    def to_voxel51_dataset(self, **kwargs) -> "fo.Dataset":
        """
        Creates a voxel51 dataset that can be used with `fo.launch_app()` to visualize it

        Kwargs:
            name (str): name of the dataset (by default uses the same name as the datasource)
            force_download (bool): download the dataset even if the size of the files is bigger than 100MB
            files_location (str|PathLike): path to the location where to download the local files
                default: ~/dagshub_datasets/user/repo/ds_name/
            redownload (bool): Redownload files, replacing the ones that might exist on the filesystem
            voxel_annotations (List[str]) : List of columns from which to load voxel annotations serialized with
                                        `to_json()`. This will override the labelstudio annotations
        """
        return self.all().to_voxel51_dataset(**kwargs)

    @property
    def default_dataset_location(self) -> Path:
        return Path(
            sanitize_filepath(os.path.join(Path.home(), "dagshub", "datasets", self.source.repo, str(self.source.id))))

    def visualize(self, **kwargs) -> "fo.Session":
        return self.all().visualize(**kwargs)

    @property
    def fields(self) -> List[MetadataFieldSchema]:
        return self.source.metadata_fields

    def annotate(self) -> Optional[str]:
        """
        Sends all datapoints in the datasource for annotation in Label Studio.
        It's recommended to not send a huge amount of datapoints to be annotated at once, to avoid overloading
        The Label Studio workspace.

        :return: Link to open Label Studio in the browser
        """
        return self.all().annotate()

    def send_to_annotation(self):
        """
        deprecated, see annotate()
        """
        return self.annotate()

    def send_datapoints_to_annotation(self, datapoints: Union[List[Datapoint], "QueryResult", List[Dict]],
                                      open_project=True, ignore_warning=False) -> Optional[str]:
        """
        Sends datapoints to annotations in Label Studio

        :param datapoints: Either a list of Datapoints or dicts that have "id" and "downloadurl" fields.
                     A QueryResult can also function as a list of Datapoint.
        :param open_project: Specifies whether the link to the returned LS project should be opened from Python
        :param ignore_warning: Suppress any non-lethal warnings that require user input
        :return: Link to open Label Studio in the browser
        """
        if len(datapoints) == 0:
            logger.warning("No datapoints provided to be sent to annotation")
            return None
        elif len(datapoints) > dagshub.common.config.recommended_annotate_limit and not ignore_warning:
            force = prompt_user(f"You are attempting to annotate {len(datapoints)} datapoints at once - it's "
                                f"recommended to only annotate up to "
                                f"{dagshub.common.config.recommended_annotate_limit} "
                                f"datapoints at a time.")
            if not force:
                return ""

        req_data = {
            "datasource_id": self.source.id,
            "datapoints": []
        }

        for dp in datapoints:
            req_dict = {}
            if type(dp) is dict:
                req_dict["id"] = dp["datapoint_id"]
                req_dict["download_url"] = dp["download_url"]
            else:
                req_dict["id"] = dp.datapoint_id
                req_dict["download_url"] = dp.download_url
            req_data["datapoints"].append(req_dict)

        init_url = multi_urljoin(self.source.repoApi.data_engine_url, "annotations/init")
        resp = http_request("POST", init_url, json=req_data, auth=self.source.repoApi.auth)

        if resp.status_code != 200:
            logger.error(f"Error while sending request for annotation: {resp.content}")
            return None
        link = resp.json()["link"]

        # Do a raw print so it works in colab/jupyter
        print("Open the following link to start working on your annotation project:")
        print(link)

        if open_project:
            webbrowser.open_new_tab(link)
        return link

    def _launch_annotation_workspace(self):
        try:
            start_workspace_url = multi_urljoin(self.source.repoApi.annotations_url, "start")
            http_request("POST", start_workspace_url, auth=self.source.repoApi.auth)
        except:  # noqa
            pass

    def wait_until_ready(self, max_wait_time=300, fail_on_timeout=True):
        """
       Blocks until the datasource preprocessing is complete

       Args:
           max_wait_time (int): Maximum time to wait in seconds
           fail_on_timeout: Whether to raise a RuntimeError or continue if the scan does not complete on time
       """

        # Start LS workspace to save time later in the flow
        self._launch_annotation_workspace()

        start = time.time()
        if max_wait_time:
            rich_console.log(f"Maximum waiting time set to {int(max_wait_time / 60)} minutes")
        spinner = rich_console.status("Waiting for datasource preprocessing to complete...")
        with spinner:
            while True:
                self.source.get_from_dagshub()
                if self.source.preprocessing_status == PreprocessingStatus.READY:
                    return

                if self.source.preprocessing_status == PreprocessingStatus.FAILED:
                    raise RuntimeError("Datasource preprocessing failed")

                if max_wait_time is not None and (time.time() - start) > max_wait_time:
                    if fail_on_timeout:
                        raise RuntimeError(
                            f"Time limit of {max_wait_time} seconds reached before processing was completed.")
                    else:
                        logger.warning(
                            f"Time limit of {max_wait_time} seconds reached before processing was completed.")
                        return

                time.sleep(1)

    def has_field(self, field_name: str):
        reserved_searchable_fields = ["path"]
        fields = (f.name for f in self.fields)
        return field_name in reserved_searchable_fields or field_name in fields

    def __repr__(self):
        res = f"Datasource {self.source.name}"
        res += f"\n\tRepo: {self.source.repo}, path: {self.source.path}"
        res += f"\n\t{self._query}"
        res += "\n\tFields:"
        for f in self.fields:
            res += f"\n\t\t{f}"
        return res + "\n"

    """ FUNCTIONS RELATED TO QUERYING
    These are functions that overload operators on the DataSet, so you can do pandas-like filtering
        ds = Dataset(...)
        queried_ds = ds[ds["value"] == 5]
    """

    def __getitem__(self, other: Union[slice, str, "Datasource"]):
        # Slicing - get items from the slice
        if type(other) is slice:
            return self.sample(other.start, other.stop)

        # Otherwise we're doing querying
        new_ds = self.__deepcopy__()
        if type(other) is str:
            if not self.has_field(other):
                raise FieldNotFoundError(other)
            new_ds._query = DatasourceQuery(other)
            return new_ds
        else:
            # "index" is a datasource with a query - compose with "and"
            # Example:
            #   ds = Dataset()
            #   filtered_ds = ds[ds["aaa"] > 5]
            #   filtered_ds2 = filtered_ds[filtered_ds["bbb"] < 4]
            if self._query.is_empty:
                new_ds._query = other._query
                return new_ds
            else:
                return other.__and__(self)

    def __gt__(self, other: object):
        self._test_not_comparing_other_ds(other)
        if not isinstance(other, (int, float, str)):
            raise NotImplementedError
        return self.add_query_op("gt", other)

    def __ge__(self, other: object):
        self._test_not_comparing_other_ds(other)
        if not isinstance(other, (int, float, str)):
            raise NotImplementedError
        return self.add_query_op("ge", other)

    def __le__(self, other: object):
        self._test_not_comparing_other_ds(other)
        if not isinstance(other, (int, float, str)):
            raise NotImplementedError
        return self.add_query_op("le", other)

    def __lt__(self, other: object):
        self._test_not_comparing_other_ds(other)
        if not isinstance(other, (int, float, str)):
            raise NotImplementedError
        return self.add_query_op("lt", other)

    def __eq__(self, other: object):
        self._test_not_comparing_other_ds(other)
        if other is None:
            return self.is_null()
        if not isinstance(other, (int, float, str)):
            raise NotImplementedError
        return self.add_query_op("eq", other)

    def __ne__(self, other: object):
        self._test_not_comparing_other_ds(other)
        if other is None:
            return self.is_not_null()
        if not isinstance(other, (int, float, str)):
            raise NotImplementedError
        return self.add_query_op("eq", other).add_query_op("not")

    def __invert__(self):
        return self.add_query_op("not")

    def __contains__(self, item):
        raise WrongOperatorError("Use `ds.contains(a)` for querying instead of `a in ds`")

    def contains(self, item: str):
        if type(item) is not str:
            return WrongOperatorError(f"Cannot use contains with non-string value {item}")
        self._test_not_comparing_other_ds(item)
        return self.add_query_op("contains", item)

    def is_null(self):
        field = self._get_filtering_field()
        value_type = _metadataTypeLookupReverse[field.valueType.value]
        return self.add_query_op("isnull", value_type())

    def is_not_null(self):
        return self.is_null().add_query_op("not")

    def _get_filtering_field(self) -> MetadataFieldSchema:
        field_name = self.get_query().column_filter
        if field_name is None:
            raise RuntimeError("The current query filter is not a field")
        for col in self.source.metadata_fields:
            if col.name == field_name:
                return col
        raise RuntimeError(f"Field {field_name} doesn't exist in the current uploaded metadata")

    def __and__(self, other: "Datasource"):
        return self.add_query_op("and", other)

    def __or__(self, other: "Datasource"):
        return self.add_query_op("or", other)

    # Prevent users from messing up their queries due to operator order
    # They always need to put the dataset query filters in parentheses, otherwise the binary and/or get executed before
    def __rand__(self, other):
        if type(other) is not Datasource:
            raise WrongOrderError(type(other))
        raise NotImplementedError

    def __ror__(self, other):
        if type(other) is not Datasource:
            raise WrongOrderError(type(other))
        raise NotImplementedError

    def add_query_op(self, op: str,
                     other: Optional[Union[str, int, float, "Datasource", "DatasourceQuery"]] = None) -> "Datasource":
        """
        Returns a new dataset with an added query param
        """
        new_ds = self.__deepcopy__()
        if type(other) is Datasource:
            other = other.get_query()
        new_ds._query.compose(op, other)
        return new_ds

    @staticmethod
    def _test_not_comparing_other_ds(other):
        if type(other) is Datasource:
            raise DatasetFieldComparisonError()


class MetadataContextManager:
    def __init__(self, datasource: Datasource):
        self._datasource = datasource
        self._metadata_entries: List[DatapointMetadataUpdateEntry] = []
        self._multivalue_fields = datasource._get_multivalue_fields()

    def update_metadata(self, datapoints: Union[List[str], str], metadata: Dict[str, Any]):
        if isinstance(datapoints, str):
            datapoints = [datapoints]

        field_value_types = {f.name: f.valueType for f in self._datasource.fields}

        for dp in datapoints:
            for k, v in metadata.items():
                if v is None:
                    continue
                if k in autogenerated_columns:
                    continue

                if type(v) is list:
                    if k not in self._multivalue_fields:
                        self._multivalue_fields.add(k)
                        # Promote all existing ones to multivalue
                        for e in self._metadata_entries:
                            if e.key == k:
                                e.allowMultiple = True
                    for sub_val in v:

                        value_type = field_value_types.get(k)
                        if value_type is None:
                            value_type = _metadataTypeLookup[type(sub_val)]
                            field_value_types[k] = value_type
                        # Don't override bytes if they're not bytes - probably just undownloaded values
                        if value_type == MetadataFieldType.BLOB and type(sub_val) is not bytes:
                            continue

                        if type(v) is bytes:
                            sub_val = self.wrap_bytes(sub_val)
                        self._metadata_entries.append(DatapointMetadataUpdateEntry(
                            url=dp,
                            key=k,
                            value=str(sub_val),
                            # todo: preliminary type check
                            valueType=value_type,
                            allowMultiple=k in self._multivalue_fields
                        ))

                else:

                    value_type = field_value_types.get(k)
                    if value_type is None:
                        value_type = _metadataTypeLookup[type(v)]
                        field_value_types[k] = value_type
                    # Don't override bytes if they're not bytes - probably just undownloaded values
                    if value_type == MetadataFieldType.BLOB and type(v) is not bytes:
                        continue

                    if type(v) is bytes:
                        v = self.wrap_bytes(v)
                    self._metadata_entries.append(DatapointMetadataUpdateEntry(
                        url=dp,
                        key=k,
                        value=str(v),
                        valueType=value_type,
                        # todo: preliminary type check
                        allowMultiple=k in self._multivalue_fields
                    ))

    @staticmethod
    def wrap_bytes(val: bytes) -> str:
        """
        Handles bytes values for uploading metadata
        The process is gzip -> base64
        """
        compressed = gzip.compress(val)
        return base64.b64encode(compressed).decode("utf-8")

    def get_metadata_entries(self):
        return self._metadata_entries



class WrongOrderError(Exception):
    def __init__(self, other_type: type):
        super().__init__()
        self.other_type = other_type

    def __str__(self):
        return f"Can't have a dataset to the right of {self.other_type}.\r\n" \
               f"Make sure to use parentheses to chain logical and/or operators.\r\n" \
               f"Example: `ds[(ds['col1'] > 1) & (ds['col2'] < 3)])`"


class DatasetFieldComparisonError(Exception):
    def __str__(self):
        return "Can't compare two fields in a dataset between each other.\r\n" \
               "Querying only supports comparisons with primitives (int/str/float)"


class WrongOperatorError(Exception):
    pass


class FieldNotFoundError(Exception):
    def __init__(self, field: str):
        super().__init__()
        self.field = field

    def __str__(self):
        return f"Field {self.field} does not exist on this datasource"


class DatasourceAlreadyExistsError(Exception):
    def __init__(self, datasource):
        super().__init__()
        self.datasource = datasource

    def __str__(self):
        return f"Datasource with name {self.datasource.name} already exists in repository {self.datasource.repo}"


class DatasourceNotFoundError(Exception):
    def __init__(self, datasource):
        super().__init__()
        self.datasource = datasource

    def __str__(self):
        return f"Datasource with name {self.datasource.name} or id {self.datasource.id} not found " \
               f"in repository {self.datasource.repo}"


class DatasetNotFoundError(Exception):
    def __init__(self, repo, id, name):
        super().__init__()
        self.repo = repo
        self.id = id
        self.name = name

    def __str__(self):
        return f"Dataset with name {self.name} or id {self.id} not found " \
               f"in repository {self.repo}"



import logging
from types import FunctionType
from typing import List, Union

from pathlib import Path
from multiprocessing import Pool, Process

from dagshub.common.api.repo import PathNotFoundError

logger = logging.getLogger(__name__)


class DagsHubDataset:
    def __init__(
        self,
        query_result,
        metadata_columns: List[str] = [],
        file_columns: List[str] = None,
        strategy: str = "lazy",
        tensorizers: Union[str, List[Union[str, FunctionType]]] = "auto",
        savedir: str = None,
        processes: int = 8,
        for_dataloader: bool = False,
    ):
        """
        query_result: <dagshub.data_engine.client.models.QueryResult>
        metadata_columns: columns that are returned from the metadata as part of the dataloader
        file_columns: columns with a datapoint metadata that are files
        strategy: preload|background|lazy; default: lazy
        tensorizers: auto|image|<function>
        savedir: location at which the dataset is stored
        processes: number of parallel processes that download the dataset
        """
        self.metadata_columns = metadata_columns
        self.entries = query_result.entries
        self.tensorizers = [
            lambda x: x,
        ] * (
            len(metadata_columns) + 1
        )  # prevent circular calls
        self.datasource = query_result.datasource
        self.repo = self.datasource.source.repoApi
        self.savedir = savedir or self.datasource.default_dataset_location
        self.strategy = strategy
        self.source = self.datasource.source.path.split("://")[0]

        self.datasource_root = Path(
            self.entries[0].path_in_repo.as_posix()[: -len(self.entries[0].path)]
        )
        self.processes = processes
        self.order = None
        self.file_columns = file_columns or self._get_file_columns()

        self.tensorizers = (
            self._get_tensorizers(tensorizers)
            if type(tensorizers) == str or type(tensorizers[0]) == str
            else tensorizers
        )

        strategy = strategy.lower()
        if strategy == "preload":
            self.pull()
        elif strategy == "background":
            if for_dataloader:
                return
            Process(target=self.pull).start()
        elif strategy != "lazy":
            logger.warning(
                "Invalid download strategy (none from preload|background|lazy); defaulting to lazy."
            )

    def __len__(self) -> int:
        return len(self.entries)

    def _get_file_columns(self):
        logger.warning("Manually detecting file columns; this may take a second.")

        res = []
        for column, value in zip(
            self.metadata_columns,
            [self.entries[0].metadata[col] for col in self.metadata_columns],
        ):
            try:
                self.repo.list_path((self.datasource_root / str(value)).as_posix())
                res.append(column)
            except PathNotFoundError:
                pass
        return res

    def get(self, idx: int) -> list:
        out = []
        entry = self.entries[idx]

        self._download(entry)
        out.append((self.savedir / entry.path).as_posix())
        for idx, column in enumerate(self.metadata_columns):
            out.append(
                (self.savedir / entry.metadata[column]).as_posix()
                if column in self.file_columns
                else entry.metadata[column]
            )

        return out

    def __getitem__(
        self, idx: int
    ) -> List[Union["torch.Tensor", "tf.Tensor"]]:  # noqa: F821
        return [
            tensorizer(data)
            for tensorizer, data in zip(self.tensorizers, self.get(idx))
        ]

    def pull(self) -> None:
        if self.order is not None:
            entries = [self.entries[idx] for idx in self.order]
            self.order = None
        else:
            entries = self.entries

        with Pool(processes=self.processes) as p:
            p.map(self._download, entries, 1)
        logger.info("Dataset download complete!")

    def _download(self, datapoint) -> None:
        paths = [
            datapoint.path,
            *[datapoint.metadata.get(column) for column in self.file_columns],
        ]

        for path in paths:
            (self.savedir / Path(path).parent).mkdir(parents=True, exist_ok=True)
            if not (self.savedir / path).is_file():
                if self.source == "repo":
                    data = self.repo.get_file(f"{self.datasource_root}/{path}")
                else:
                    data = self.repo.get_storage_file(
                        f"{'/'.join(list(self.datasource.source.path_parts().values())[:2])}"
                        f"/{self.datasource_root}/{path}"
                    )

                filepath = self.savedir / path
                with open(filepath, "wb") as file:
                    file.write(data)

    def _get_tensorizers(
        self, datatypes: Union[str, List[Union[str, FunctionType]]]
    ) -> FunctionType:
        if datatypes in ["auto", "guess"]:  # guess is an easter egg argument
            logger.warning("`tensorizers` set to 'auto'; guessing the datatypes")
            tensorizers = []

            # naive pass
            for idx, entry in enumerate(self.get(0)):
                if not idx or self.metadata_columns[idx - 1] in self.file_columns:
                    extension = entry.split("/")[-1].split(".")[-1]
                    if extension in ["mkv", "mp4"]:
                        tensorizers.append(self.tensorlib.video)
                    elif extension in ["wav", "mp3"]:
                        tensorizers.append(self.tensorlib.audio)
                    elif extension in ["png", "jpg", "jpeg"]:
                        tensorizers.append(self.tensorlib.image)
                    else:
                        raise ValueError(
                            "Unable to automatically detect the datatypes. "
                            "Please manually set a list of tensorizers, either with string arguments image|video|audio,"
                            " or custom tensorizer functions with prototype `<str> -> <torch.Tensor>`."
                        )
                elif type(entry) in [int, float]:
                    tensorizers.append(self.tensorlib.numeric)
                else:
                    raise ValueError(
                        "Unable to automatically tensorize non-numeric metadata. "
                        "Please manually setup a list of tensorizers, either with string arguments image|video|audio, "
                        "or custom tensorizer functions with prototype `<str> -> <torch.Tensor>`."
                    )
            return tensorizers
        elif datatypes in ["image", "audio", "video"]:
            return [
                getattr(self.tensorlib, datatypes),
            ] * (len(self.metadata_columns) + 1)
        elif (
            type(datatypes) == list and len(datatypes) == len(self.metadata_columns) + 1
        ):
            return [
                getattr(self.tensorlib, datatype) if type(datatype) == str else datatype
                for datatype in datatypes
            ]
        else:
            raise ValueError(
                "Unable to set tensorizers. "
                "Please ensure the number of selected columns equals the number of tensorizers."
            )



import os
import shutil
import sys
from typing import TYPE_CHECKING

from dagshub.common.helpers import prompt_user

if TYPE_CHECKING:
    from starlette.requests import Request
    from dagshub.data_engine.voxel_plugin_server.models import PluginServerState


def set_voxel_envvars():
    """
    Sets the environment variables relevant for voxel:
        FIFTYONE_PLUGINS_DIR - where to load the plugins from
    """
    script_dir = os.path.dirname(__file__)
    plugin_dir = os.environ.get("DAGSHUB_FIFTYONE_PLUGINS_DIR",
                                os.path.join(script_dir, "plugins"))
    plugin_dir = os.path.abspath(plugin_dir)
    fo_dir_envkey = "FIFTYONE_PLUGINS_DIR"
    if fo_dir_envkey in os.environ and os.environ[fo_dir_envkey] != plugin_dir:
        plugins_dest_dir = os.path.join(os.environ[fo_dir_envkey], "dagshub")
        # TODO: handle version changes, for now prompt only if the plugin isn't copied
        if not os.path.exists(plugins_dest_dir):
            response = prompt_user(f"You have {fo_dir_envkey} env var setup to {os.environ[fo_dir_envkey]}. "
                                   "Do you want to copy the dagshub plugin there?"
                                   "You need the plugin copied in order to get full integration experience",
                                   default=True)
            if not response:
                return
            shutil.copytree(os.path.join(plugin_dir, "dagshub"), plugins_dest_dir)
    else:
        os.environ[fo_dir_envkey] = plugin_dir

    # Reset the plugin dir if voxel has already been imported
    if "fiftyone" in sys.modules:
        import fiftyone as fo
        fo.config.plugins_dir = plugin_dir


def get_plugin_state(request: "Request") -> "PluginServerState":
    return request.app.state.PLUGIN_STATE



import datetime
import logging
import os
import traceback
from typing import Optional, Dict, List

import yaml

from dagshub.auth import oauth
from dagshub.common import config

logger = logging.getLogger(__name__)

APP_TOKEN_TYPE = "app-token"


class TokenStorage:
    def __init__(self, cache_location: str = None, **kwargs):
        cache_location = cache_location or config.cache_location
        self.cache_location = cache_location
        self.__token_cache: Optional[Dict[str, List[Dict]]] = None

    @property
    def _token_cache(self):
        if self.__token_cache is None:
            self.__token_cache = self._load_cache_file()
        return self.__token_cache

    def add_token(self, token: Dict, host: str = None):
        host = host or config.host
        if host not in self._token_cache:
            self._token_cache[host] = []
        self._token_cache[host].append(token)
        self._store_cache_file()

    def get_token(self, host: str = None, fail_if_no_token: bool = False, **kwargs):
        host = host or config.host
        tokens = self._token_cache.get(host, [])
        app_tokens = [t for t in tokens if t.get("token_type") == APP_TOKEN_TYPE]
        if len(app_tokens) > 0:
            token = app_tokens[0]
        else:
            non_expired_tokens = [t for t in tokens if not self._is_expired(t)]
            if len(non_expired_tokens) > 0:
                token = non_expired_tokens[0]
            elif fail_if_no_token:
                raise RuntimeError(
                    f"No valid tokens found for host '{host}'.\n"
                    "Log into DagsHub by executing `dagshub login` in your terminal")
            else:
                logger.debug(
                    f"No valid tokens found for host '{host}'. Authenticating with OAuth"
                )
                token = oauth.oauth_flow(host, **kwargs)
                tokens.append(token)
                self._token_cache[host] = tokens
                self._store_cache_file()
        return token["access_token"]

    @staticmethod
    def _is_expired(token: Dict[str, str]) -> bool:
        if "expiry" not in token:
            return True
        if token["expiry"] == "never":
            return False
        # Need to cut off the three additional precision numbers in milliseconds, because %f only parses 6 digits
        expiry = token["expiry"][:-4] + "Z"
        expiry_dt = datetime.datetime.strptime(expiry, "%Y-%m-%dT%H:%M:%S.%fZ")
        is_expired = expiry_dt < datetime.datetime.utcnow()
        return is_expired

    def _load_cache_file(self) -> Dict[str, List[Dict]]:
        logger.debug(f"Loading OAuth token cache from {self.cache_location}")
        if not os.path.exists(self.cache_location):
            logger.debug("OAuth token cache file doesn't exist")
            return self._get_empty_cache_dict()
        try:
            with open(self.cache_location) as f:
                tokens_cache = yaml.load(f, yaml.Loader)
                return tokens_cache
        except Exception:
            logger.error(
                f"Error while loading DagsHub OAuth token cache: {traceback.format_exc()}"
            )
            raise

    @staticmethod
    def _get_empty_cache_dict():
        return {"version": config.TOKENS_CACHE_SCHEMA_VERSION}

    def _store_cache_file(self):
        logger.debug(f"Dumping OAuth token cache to {self.cache_location}")
        try:
            dirpath = os.path.dirname(self.cache_location)
            if not os.path.exists(dirpath):
                os.makedirs(dirpath)
            with open(self.cache_location, "w") as f:
                yaml.dump(self.__token_cache, f, yaml.Dumper)
        except Exception:
            logger.error(
                f"Error while storing DagsHub OAuth token cache: {traceback.format_exc()}"
            )
            raise


_token_storage: Optional[TokenStorage] = None


def _get_token_storage(**kwargs):
    global _token_storage
    if _token_storage is None:
        _token_storage = TokenStorage(**kwargs)
    return _token_storage


def get_token(**kwargs):
    """
    Gets a DagsHub token, by default if no token is found authenticates with OAuth

    Kwargs:
        host (str): URL of a dagshub instance (defaults to dagshub.com)
        cache_location (str): Location of the cache file with the token (defaults to <cache_dir>/dagshub/tokens)
        fail_if_no_token (bool): What to do if token is not found.
            If set to False (default), goes through OAuth flow
            If set to True, throws a RuntimeError
    """
    return _get_token_storage(**kwargs).get_token(**kwargs)


def add_app_token(token: str, host: Optional[str] = None, **kwargs):
    token_dict = {
        "access_token": token,
        "token_type": APP_TOKEN_TYPE,
        "expiry": "never",
    }
    _get_token_storage(**kwargs).add_token(token_dict, host)


def add_oauth_token(host: Optional[str] = None, **kwargs):
    host = host or config.host
    token = oauth.oauth_flow(host)
    _get_token_storage(**kwargs).add_token(token, host)



import logging
from typing import Optional, Union, List

from dagshub.common.analytics import send_analytics_event
from dagshub.common.api.repo import RepoAPI
from dagshub.data_engine.client.data_client import DataClient
from dagshub.data_engine.model.datasource import Datasource
from dagshub.data_engine.model.datasource_state import DatasourceState, DatasourceType, path_regexes
from dagshub.data_engine.model.errors import DatasourceNotFoundError

logger = logging.getLogger(__name__)


def create_datasource(repo: str, name: str, path: str, revision: Optional[str] = None) -> Datasource:
    """
    Create a datasource from a path in the repo, or a storage bucket URL.

    :param repo: The full repo name in <owner>/<reponame> format
    :param name: The name of the datasource to be created
    :param path: Either a path to a directory inside the Git/DVC repo on DagsHub, e.g. "path/to/dir"
                 or else a URL pointing
                 to a storage bucket which is connected to the DagsHub repo - e.g. "s3://bucketname/path/in/bucket"
    :param revision: Only valid when using a Git/DVC path inside the DagsHub repo, not when using a bucket URL.
                     Allows you to define the branch name where the referenced directory exists.
                     The default repo branch is used if this is left blank.
    :return The created datasource
    """

    if path_regexes[DatasourceType.BUCKET].fullmatch(path):
        if revision is not None:
            raise ValueError("revision cannot be used together with bucket URLs")
        return create_from_bucket(repo, name, bucket_url=path)
    else:
        return create_from_repo(repo, name, path=path, revision=revision)


create = create_datasource


def get_or_create(repo: str, name: str, path: str, revision: Optional[str] = None) -> Datasource:
    """
    First attempts to get the repo datasource with the given name, and only if that fails,
    invokes create_datasource with the given parameters.
    See the docs on create_datasource for more info.
    """
    try:
        return get_datasource(repo, name)
    except DatasourceNotFoundError:
        return create_datasource(repo, name, path, revision)


def create_from_bucket(repo: str, name: str, bucket_url: str) -> Datasource:
    # TODO: validation
    source = _create_datasource_state(repo, name, DatasourceType.BUCKET, bucket_url)
    return Datasource(source)


def create_from_repo(repo: str, name: str, path: str, revision: Optional[str] = None) -> Datasource:
    if revision is None:
        repo_api = RepoAPI(repo)
        revision = repo_api.default_branch
    url = f"repo://{repo}/{revision}:{path.lstrip('/')}"
    source = _create_datasource_state(repo, name, DatasourceType.REPOSITORY, url)
    if revision is not None:
        source.revision = revision
    return Datasource(source)


def get_datasource(repo: str, name: Optional[str] = None, id: Optional[Union[int, str]] = None, **kwargs) -> Datasource:
    """
    Additional kwargs:
    revision - for repo datasources defines which branch/revision to download from (default branch if not specified)
    """
    ds_state = DatasourceState(repo=repo, name=name, id=id)
    ds_state.get_from_dagshub()
    if "revision" in kwargs:
        ds_state.revision = kwargs["revision"]
    return Datasource(ds_state)


def get_datasources(repo: str) -> List[Datasource]:
    send_analytics_event("Client_DataEngine_getDatasources")
    client = DataClient(repo)
    sources = client.get_datasources(None, None)
    return [Datasource(DatasourceState.from_gql_result(repo, source)) for source in sources]


# Alias
get = get_datasource


def _create_datasource_state(repo: str, name: str, source_type: DatasourceType, path: str) -> DatasourceState:
    ds = DatasourceState(name=name, repo=repo)
    ds.source_type = source_type
    ds.path = path
    ds.create()
    return ds


__all__ = [
    create_datasource,
    create,
    create_from_bucket,
    create_from_repo,
    get_datasource,
    get,
    get_or_create,
]



from typing import Optional, TYPE_CHECKING, Any, List

from dataclasses import dataclass

from dagshub.data_engine.model.datasource import Datasource
from dagshub.data_engine.model.query import DatasourceQuery

if TYPE_CHECKING:
    import fiftyone as fo


@dataclass
class PluginServerState:
    voxel_session: "fo.Session"
    datasource: "Datasource"
    branch: Optional[str] = None


@dataclass
class LabelStudioProject:
    id: int
    name: str
    branch: str


@dataclass
class VoxelFilterState:
    _CLS: str
    exclude: bool
    isMatching: bool
    range: Optional[List[Any]]
    values: Optional[List[Any]]
    filter_field: Optional[str]

    def to_datasource_query(self) -> DatasourceQuery:
        # TODO: handle exclude and isMatching
        if self.range is not None:
            # range: field >= min AND field <= max
            resQuery = DatasourceQuery()
            if self.range[0] is not None:
                q1 = DatasourceQuery(self.filter_field)
                q1.compose("ge", self.range[0])
                resQuery.compose("and", q1)
            if self.range[1] is not None:
                q2 = DatasourceQuery(self.filter_field)
                q2.compose("le", self.range[1])
                resQuery.compose("and", q2)
            return resQuery

        if self.values is not None:
            #
            res_q = DatasourceQuery()
            for val in self.values:
                q = DatasourceQuery(self.filter_field)
                q.compose("eq", val)
                res_q.compose("or", q)
            return res_q

        raise NotImplementedError
